"""Integration for `Pandas <https://pandas.pydata.org/>`_ types."""

import typing as _t
import warnings as _warnings
from collections import abc as _abc
from collections import defaultdict as _defaultdict

import pandas as _pd

from id_translation import dio as _dio
from id_translation import types as _tt
from id_translation.dio.default import _sequence
from id_translation.dio.exceptions import NotInplaceTranslatableError as _NotInplaceTranslatableError
from id_translation.offline import TranslationMap as _TranslationMap

PandasT = _t.TypeVar("PandasT", _pd.DataFrame, _pd.Series, _pd.Index, _pd.MultiIndex)
"""Supported ``pandas`` types."""
_VectorT = _t.TypeVar("_VectorT", _pd.Series, _pd.Index)
"""Types with ``ndim==1``."""


class PandasIO(_dio.DataStructureIO[PandasT, _tt.NameType, _tt.SourceType, _tt.IdType]):
    """Optional IO implementation for ``pandas`` types."""

    priority = 1999

    @classmethod
    def handles_type(cls, arg: _t.Any) -> bool:
        return isinstance(arg, (_pd.DataFrame, _pd.Series, _pd.Index))

    @classmethod
    def names(cls, translatable: PandasT) -> list[_tt.NameType] | None:
        if isinstance(translatable, _pd.DataFrame):
            columns = translatable.columns
            if isinstance(columns, _pd.MultiIndex):
                return list(columns.get_level_values(-1))
            else:
                return list(columns)

        if isinstance(translatable, _pd.MultiIndex):
            names = [n for n in translatable.names if n is not None]
            return names or None

        return None if translatable.name is None else [translatable.name]

    @classmethod
    def extract(cls, translatable: PandasT, names: list[_tt.NameType]) -> dict[_tt.NameType, _abc.Sequence[_tt.IdType]]:
        if isinstance(translatable, _pd.DataFrame):
            ans = _defaultdict(list)
            for i, name in enumerate(translatable.columns):
                if name in names:
                    ans[name].extend(_float_to_int(translatable.iloc[:, i]))
            return dict(ans)
        elif isinstance(translatable, _pd.MultiIndex):
            # This will error on duplicate names, which is probably a good thing.
            return {name: list(translatable.unique(name)) for name in names}
        elif isinstance(translatable, (_pd.Series, _pd.Index)):
            _sequence.verify_names(len(translatable), names)
            translatable = _float_to_int(translatable)
            if len(names) == 1:
                return {names[0]: translatable.unique().tolist()}
            else:
                return _sequence.SequenceIO.extract(translatable, names)

        raise TypeError(f"This should not happen: {type(translatable)=}")  # pragma: no cover

    @classmethod
    def insert(
        cls,
        translatable: PandasT,
        names: list[_tt.NameType],
        tmap: _TranslationMap[_tt.NameType, _tt.SourceType, _tt.IdType],
        copy: bool,
    ) -> PandasT | None:
        if not copy:
            if isinstance(translatable, _pd.DataFrame):
                pass  # Ok
            elif isinstance(translatable, _pd.Series):
                pass  # Ok, for now.   # TODO(issues/170): PDEP-6 check
            else:
                raise _NotInplaceTranslatableError(translatable)

        if isinstance(translatable, _pd.Index):
            return _translate_index(translatable, names, tmap)

        if isinstance(translatable, _pd.DataFrame):
            return _translate_frame(translatable, names, tmap, copy)

        if isinstance(translatable, _pd.Series):
            return _translate_series(translatable, names, tmap, copy)

        raise TypeError(f"This should not happen: {type(translatable)=}")  # pragma: no cover


def _translate_pandas_vector(
    pvt: _VectorT,
    names: list[_tt.NameType],
    tmap: _TranslationMap[_tt.NameType, _tt.SourceType, _tt.IdType],
) -> list[str | None] | _VectorT:
    _sequence.verify_names(len(pvt), names)

    if len(names) == 1:
        # Optimization for single-name vectors. Faster than SequenceIO for pretty much every size.
        magic_dict = tmap[names[0]]

        mapping: dict[_tt.IdType, str | None]
        if _pd.api.types.is_numeric_dtype(pvt):
            # We don't need to cast float to int here, since hash(1.0) == hash(1). The cast in extract() is required
            # because some database drivers may complain, especially if they receive floats (especially NaN).
            mapping = {idx: magic_dict[idx] for idx in pvt.unique()}
            return pvt.map(mapping)
        else:
            mapping = {}
            data: list[_t.Any] = pvt.to_list()
            for i, idx in enumerate(data):
                if idx in mapping:
                    value = mapping[idx]
                else:
                    value = magic_dict[idx]
                    mapping[idx] = value
                data[i] = value

        return data
    else:
        return _sequence.translate_sequence(pvt, names, tmap)


def _translate_index(
    index: PandasT,
    names: list[_tt.NameType],
    tmap: _TranslationMap[_tt.NameType, _tt.SourceType, _tt.IdType],
) -> PandasT | None:
    if isinstance(index, _pd.MultiIndex):
        df = index.to_frame()
        _translate_frame(df, names, tmap, copy=False)
        return _pd.MultiIndex.from_frame(df, names=index.names)

    result = _translate_pandas_vector(index, names, tmap)
    if isinstance(result, _pd.Index):
        return result
    return _pd.Index(result, name=index.name, copy=False)


def _translate_frame(
    df: _pd.DataFrame,
    names: list[_tt.NameType],
    tmap: _TranslationMap[_tt.NameType, _tt.SourceType, _tt.IdType],
    copy: bool,
) -> _pd.DataFrame:
    if copy:
        df = df.copy()

    original_columns = df.columns

    try:
        df.columns = _pd.RangeIndex(len(original_columns))
        for name, int_col in zip(original_columns, df.columns, strict=True):
            if name in names:
                translated = _translate_pandas_vector(df[int_col], [name], tmap)
                df[int_col] = translated
    finally:
        df.columns = original_columns

    return df if copy else None


def _translate_series(
    series: _pd.Series,
    names: list[_tt.NameType],
    tmap: _TranslationMap[_tt.NameType, _tt.SourceType, _tt.IdType],
    copy: bool,
) -> _pd.Series | None:
    result = _translate_pandas_vector(series, names, tmap)
    if copy:
        if isinstance(result, _pd.Series):
            return result
        return _pd.Series(result, index=series.index, name=series.name, copy=False)

    with _warnings.catch_warnings():
        _warnings.simplefilter(action="ignore", category=FutureWarning)  # TODO(issues/170): PDEP-6 check
        series[:] = result
    return None


def _float_to_int(pvt: _VectorT) -> _VectorT:
    return pvt.dropna().astype(int) if _pd.api.types.is_float_dtype(pvt) else pvt



"""Classes used for ID and translation transformation."""

import typing as _t

from ..types import IdType as _IdType
from ..types import SourceType as _SourceType


@_t.runtime_checkable
class Transformer(_t.Protocol[_IdType]):
    """Transformation API type.

    Transformers are persistent entities owned by a single :class:`.Translator` instance. See the
    :class:`.BitmaskTransformer` for a concrete example.
    """

    def update_ids(self, ids: set[_IdType], /) -> None:
        """Transform a source-to-ids mapping dict.

        Called just before IDs are fetched from the source.

        .. note::

           Not called when the :class:`.Translator` is working offline.

        Args:
            ids: A collection of IDs from the `source`.
        """

    def update_translations(self, translations: dict[_IdType, str], /) -> None:
        """Update real translations.

        Called by the :class:`.MagicDict` during initialization.

        Args:
            translations: A dict of real translations ``{id: translation_string}``.
        """

    def try_add_missing_key(
        self,
        key: _IdType,
        /,
        *,
        translations: _t.MutableMapping[_IdType, str],
    ) -> None:
        """Attempt to create and add a translation for an unknown ID.

        Callback function used by :class:`.MagicDict` whenever an unknown ID is requested.

        Args:
            key: An ID which is not present in the `translations`.
            translations: A mutable mapping of translations. Typically, the :class:`.MagicDict` caller itself.
        """


Transformers = dict[_SourceType, Transformer[_IdType]]
"""A dict ``{source: transformer}`` of initialized :class:`.Transformer` instances."""



from typing import Any

from id_translation.fetching import AbstractFetcher
from id_translation.types import IdType, SourceType

from ._initialize import initialize


def default_fetcher_factory(clazz: str, config: dict[str, Any]) -> AbstractFetcher[SourceType, IdType]:
    """Create an :class:`.AbstractFetcher` from config."""
    from id_translation import fetching as default_module  # noqa: PLC0415

    return initialize(
        clazz,
        config,
        AbstractFetcher,  # type: ignore[type-abstract]  # https://github.com/python/mypy/issues/4717
        default_module,
    )



import logging
from collections.abc import Callable, Generator, Iterable
from contextlib import contextmanager
from os import getenv
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, TypeAlias

from rics.collections.dicts import InheritedKeysDict
from rics.env.read import read_bool
from rics.types import AnyPath

from id_translation.exceptions import ConfigurationError
from id_translation.fetching import AbstractFetcher, CacheAccess, Fetcher, MultiFetcher
from id_translation.mapping import Mapper
from id_translation.transform.types import Transformer, Transformers
from id_translation.types import IdType, NameType, SourceType

from . import factories as cf
from ._load_toml import load_toml_file
from .meta import ConfigMetadata, Metaconf

if TYPE_CHECKING:
    from id_translation import Translator


SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS = "ID_TRANSLATION_SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS"


class TranslatorFactory(Generic[NameType, SourceType, IdType]):
    """Create a :class:`.Translator` from TOML inputs."""

    FetcherFactory: TypeAlias = Callable[[str, dict[str, Any]], AbstractFetcher[Any, Any]]
    """Signature for :attr:`FETCHER_FACTORY`."""

    FETCHER_FACTORY: FetcherFactory = staticmethod(cf.default_fetcher_factory)
    """A callable ``(clazz, config) -> AbstractFetcher``.

    Overwrite attribute with your own :attr:`.FetcherFactory` implementation to customize.

    Args:
        clazz: Type of :class:`.AbstractFetcher` to create.
        config: Keyword arguments for the fetcher class.

    Returns:
        An :class:`.AbstractFetcher` instance.

    Raises:
        exceptions.ConfigurationError: If `config` is invalid.
        TypeError: If `clazz` is not an :class:`.AbstractFetcher` subtype.

    See Also:
        :ref:`translator-config-fetching`
    """

    MapperFactory: TypeAlias = Callable[[dict[str, Any], bool], Mapper[Any, Any, Any] | None]
    """Signature for :attr:`MAPPER_FACTORY`."""

    MAPPER_FACTORY: MapperFactory = cf.default_mapper_factory
    """A callable ``(config, for_fetcher) -> Mapper | None``.

    Overwrite attribute with your own :attr:`.MapperFactory` implementation to customize.

    If ``None`` is returned, a suitable default is used instead.

    Args:
        config: Keyword arguments for the :class:`.Mapper`.
        for_fetcher: Flag indicating that the :class:`.Mapper` returned will be used by an :class:`.AbstractFetcher` instance.

    Returns:
        A :class:`.Mapper` instance or ``None``.

    Raises:
        ConfigurationError: If `config` is invalid.

    See Also:
        :ref:`translator-config-mapping`
    """

    TransformerFactory: TypeAlias = Callable[[str, dict[str, Any]], Transformer[Any]]
    """Signature for :attr:`TRANSFORMER_FACTORY`."""

    TRANSFORMER_FACTORY: TransformerFactory = cf.default_transformer_factory
    """A callable ``(clazz, config) -> Transformer``.

    Overwrite attribute with your own :attr:`.TransformerFactory` implementation to customize.

    Args:
        clazz: Type of :class:`.Transformer` to create.
        config: Keyword arguments for the transformer class.

    Returns:
        A :class:`.Transformer` instance.

    Raises:
        ConfigurationError: If `config` is invalid.

    See Also:
        :ref:`translator-config-transform`
    """

    CacheAccessFactory: TypeAlias = Callable[[str, dict[str, Any]], CacheAccess[Any, Any]]
    """Signature for :attr:`CACHE_ACCESS_FACTORY`."""

    CACHE_ACCESS_FACTORY: CacheAccessFactory = cf.default_cache_access_factory
    """A callable ``(clazz, config) -> CacheAccess``.

    Overwrite attribute with your own :attr:`.CacheAccessFactory` implementation to customize.

    Args:
        clazz: Type of :class:`.CacheAccess` to create.
        config: Keyword arguments for the cache class.

    Returns:
        A :class:`.CacheAccess` instance.

    Raises:
        ConfigurationError: If `config` is invalid.
    """

    TOP_LEVEL_KEYS = ("translator", "mapping", "fetching", "unknown_ids", "transform")
    """Top-level keys allowed in the main configuration file."""

    def __init__(
        self,
        file: AnyPath,
        fetchers: Iterable[AnyPath],
        clazz: type["Translator[NameType, SourceType, IdType]"] | None = None,
        suppress_optional_fetcher_init_errors: bool | None = None,
    ) -> None:
        from id_translation import Translator  # noqa: PLC0415

        self.file = str(file)
        self.extra_fetchers = list(map(str, fetchers))
        self.clazz: type[Translator[NameType, SourceType, IdType]] = clazz or Translator[NameType, SourceType, IdType]

        metaconf_path = Path(self.file).with_name("metaconf.toml")
        self._metaconf = Metaconf.from_path_or_default(metaconf_path)
        self.logger = logging.getLogger(__package__).getChild(type(self).__name__)

        if suppress_optional_fetcher_init_errors is None:
            suppress_optional_fetcher_init_errors = read_bool(SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS)
        self.suppress_optional_fetcher_init_errors = suppress_optional_fetcher_init_errors

    @property
    def metaconf(self) -> Metaconf:
        """Returns the :class:`.Metaconf` used by this factory."""
        return self._metaconf

    def create(self) -> "Translator[NameType, SourceType, IdType]":
        """Create :class:`.Translator` instance."""
        config_metadata = ConfigMetadata.from_toml_paths(self.file, self.extra_fetchers, self.clazz)
        with _rethrow_with_file(self.file):
            config: dict[str, Any] = self.load_toml_file(self.file)

        fetcher, fetcher_transformers = self._handle_fetching(
            config.pop("fetching", {}), self.extra_fetchers, _identifier_from_config_metadata(config_metadata)
        )

        with _rethrow_with_file(self.file):
            _check_allowed_keys(self.TOP_LEVEL_KEYS, actual=config, toml_path="<root>")
            translator_config = config.pop("translator", {})
            mapper = self._make_mapper("translator", translator_config)
            _make_default_translations(translator_config, config.pop("unknown_ids", {}))

            translator_transformers = self._handler_transformers(config.pop("transform", {}))
            if keys := set(fetcher_transformers).intersection(translator_transformers):
                msg = f"Transformers for {len(keys)} sources also defined on the fetcher level: {keys}."
                raise ValueError(msg)
            translator_config["transformers"] = translator_transformers | fetcher_transformers

            ans = self.clazz(
                fetcher,
                mapper=mapper,
                **translator_config,
            )

            ans._config_metadata = config_metadata
            return ans

    def load_toml_file(self, path: str) -> dict[str, Any]:
        """Read a TOML file from `path` with the current :attr:`.Metaconf.env` settings.

        Args:
            path: Path to file.

        Returns:
            A dict parsed from `path`.

        See Also:
            :func:`.load_toml_file`
        """
        env = self.metaconf.env
        return load_toml_file(
            path,
            allow_interpolation=env.allow_interpolation,
            allow_nested=env.allow_nested,
            allow_blank=env.allow_blank,
        )

    def _handle_fetching(
        self,
        config: dict[str, Any],
        extra_fetchers: list[str],
        default_identifiers: list[list[str]],
    ) -> tuple[Fetcher[SourceType, IdType], Transformers[SourceType, IdType]]:
        multi_fetcher_kwargs = config.pop("MultiFetcher", {})

        fetchers: list[Fetcher[SourceType, IdType]] = []
        transformers: Transformers[SourceType, IdType] = {}

        if config:
            with _rethrow_with_file(self.file, show_init_errors_hint=True):
                fetcher = self._make_fetcher(default_identifiers[0], **config)
                if isinstance(fetcher, Exception):
                    self._log_optional_fetcher_init_error(fetcher, str(self.file))
                else:
                    fetchers.append(fetcher)  # Add primary fetcher

        for i, fetcher_file in enumerate(extra_fetchers, start=1):
            with _rethrow_with_file(fetcher_file, show_init_errors_hint=True):
                fetcher_config = self.load_toml_file(fetcher_file)
                _check_allowed_keys(["fetching", "transform"], actual=fetcher_config, toml_path="<root>")
                fetcher = self._make_fetcher(default_identifiers[i], **fetcher_config["fetching"])
                if isinstance(fetcher, Exception):
                    self._log_optional_fetcher_init_error(fetcher, fetcher_file)
                    continue
                new_transformers = self._handler_transformers(fetcher_config.get("transform", {}))

                if keys := set(new_transformers).intersection(transformers):
                    msg = f"Transformers for {len(keys)} sources were already defined in another fetcher file: {keys}."
                    raise ValueError(msg)
                transformers.update(new_transformers)
            fetchers.append(fetcher)

        if not fetchers:
            raise ConfigurationError(
                f"At least one [fetching]-section is required. Add it to '{self.file}',"
                " or as an auxiliary configuration.",
            )

        retval: Fetcher[SourceType, IdType]
        if len(fetchers) == 1:
            if multi_fetcher_kwargs and self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug(
                    f"MultiFetcher arguments {multi_fetcher_kwargs} are ignored; only one fetcher defined."
                )
            retval = fetchers[0]
        else:
            retval = MultiFetcher(*fetchers, **multi_fetcher_kwargs)
        return retval, transformers

    def _log_optional_fetcher_init_error(self, exception: BaseException, fetcher_file: str) -> None:
        value = getenv(SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS)
        env = f"{SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS}={value}"
        url = "https://id-translation.readthedocs.io/en/stable/documentation/translator-config.html#optional-fetchers"
        self.logger.exception(
            f"Discarded optional fetcher in file '{fetcher_file}': {exception!r}."
            f"\nHint: Discarded since `optional=true` and `{env}`."
            f"\nHint: See {url} for help.",
            exc_info=exception,
            extra={"fetcher_file": str(fetcher_file), "reason": str(exception)},
        )

    @classmethod
    def _make_mapper(cls, parent_section: str, config: dict[str, Any]) -> Mapper[Any, Any, Any] | None:
        if "mapping" not in config:
            return None  # pragma: no cover

        config = config.pop("mapping")
        for_fetcher = parent_section.startswith("fetching")
        if for_fetcher:
            config = {**AbstractFetcher.default_mapper_kwargs(), **config}

        return cls.MAPPER_FACTORY(config, for_fetcher)

    @classmethod
    def _make_cache_access(cls, config: dict[str, Any]) -> CacheAccess[Any, Any]:
        return cls.CACHE_ACCESS_FACTORY(config.pop("type"), config)

    def _make_fetcher(self, __identifiers: list[str], **config: Any) -> AbstractFetcher[SourceType, IdType] | Exception:
        mapper = self._make_mapper("fetching", config) if "mapping" in config else None
        cache_access = self._make_cache_access(config.pop("cache")) if "cache" in config else None

        if len(config) == 0:  # pragma: no cover
            raise ConfigurationError("Fetcher implementation section missing.")
        if len(config) > 1:  # pragma: no cover
            raise ConfigurationError(f"Multiple fetcher implementations specified in the same file: {sorted(config)}")

        clazz, kwargs = next(iter(config.items()))

        kwargs["identifiers"] = kwargs.get("identifiers", __identifiers)
        kwargs["mapper"] = mapper
        kwargs["cache_access"] = cache_access

        is_optional = kwargs.get("optional")
        if isinstance(is_optional, bool):
            # Only if ID_TRANSLATION_SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS=true.
            is_optional = is_optional and self.suppress_optional_fetcher_init_errors
        else:
            is_optional = False

        if is_optional:
            try:
                return self.FETCHER_FACTORY(clazz, kwargs)
            except Exception as e:
                return e
        else:
            return self.FETCHER_FACTORY(clazz, kwargs)

    @classmethod
    def _handler_transformers(cls, per_source: dict[SourceType, dict[str, Any]]) -> Transformers[SourceType, IdType]:
        transformers = {}

        for source, config in per_source.items():
            if len(config) != 1:
                raise ConfigurationError(
                    "Transformation config must be specified as [transform.<source>.<transformer-class>] sections."
                )
            clazz, kwargs = next(iter(config.items()))
            transformers[source] = cls.TRANSFORMER_FACTORY(clazz, kwargs)
        return transformers


def _make_default_translations(
    out: dict[str, Any],
    config: dict[str, Any],
) -> None:
    _check_allowed_keys(["fmt", "overrides"], actual=config, toml_path="translator.unknown_ids")

    if "fmt" in config:
        out["default_fmt"] = config.pop("fmt")
    if "overrides" in config:
        shared, specific = _split_overrides(config.pop("overrides"))
        out["default_fmt_placeholders"] = InheritedKeysDict(specific, default=shared)


def _check_allowed_keys(allowed: Iterable[str], *, actual: Iterable[str], toml_path: str) -> None:
    bad_keys = set(actual).difference(allowed)
    if bad_keys:
        raise ValueError(f"Forbidden keys {sorted(bad_keys)} in [{toml_path}]-section.")


def _split_overrides(overrides: Any) -> Any:
    specific = {k: v for k, v in overrides.items() if isinstance(v, dict)}
    shared = {k: v for k, v in overrides.items() if k not in specific}
    return shared, specific


def _identifier_from_config_metadata(config_metadata: ConfigMetadata) -> list[list[str]]:
    # Use the config filename and sha hash as the default keys
    return list(map(lambda t: [t[0].name, t[1]], (config_metadata.main, *config_metadata.extra_fetchers)))


@contextmanager
def _rethrow_with_file(
    file: str,
    *,
    show_init_errors_hint: bool = False,
) -> Generator[None, None, None]:
    try:
        yield
    except Exception as e:
        file_hint = f"In file: {Path(file).resolve()}"
        notes = [file_hint]
        if show_init_errors_hint:
            notes.append(f"Setting {SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS}=true may help temporarily.")

        for hint in notes:
            e.add_note(f"Hint: {hint}")

        if isinstance(e, ConfigurationError):
            raise
        else:
            msg = f"{type(e).__name__}: {e}\n    raised when parsing file: {Path(file).resolve()}"
            raise ConfigurationError(msg) from e



import logging
from collections.abc import Iterator, MutableMapping
from typing import Any
from uuid import UUID

from .. import _uuid_utils
from ..transform.types import Transformer
from ..types import IdType
from ._format import Format
from .types import TranslatedIds


class MagicDict(MutableMapping[IdType, str]):
    """Dictionary type for translated IDs.

    A ``dict``-like mapping which returns "real" values if present in a backing dict. Values for unknown keys are
    generated using the :attr:`default_value`.

    Args:
        real_translations: A dict holding :attr:`real` translations.
        default_value: A string with exactly one or zero positional placeholders.
        enable_uuid_heuristics: Improves matching when :py:class:`~uuid.UUID`-like IDs are in use. Forcibly set to
            ``False`` if any of the `real_translations` are not ``UUID``-like.
        transformer: Initialized :class:`.Transformer` instance. The :meth:`.Transformer.update_translations`-method is
            called after UUID heuristics are applied.

    Examples:
        **Similarities with the built-in dict**

        >>> magic = MagicDict({1999: "Sofia", 1991: "Richard"})

        Iteration, equality, and length are based on the :attr:`real`  values.

        >>> magic
        {1999: 'Sofia', 1991: 'Richard'}
        >>> len(magic)
        2
        >>> list(magic)
        [1999, 1991]
        >>> magic.real == magic
        True
        >>> magic == {1999: "Sofia"}  # Element missing
        False

        As you'd expect, casting to a regular ``dict`` removes all special handling.

        **Differences from the built-in dict**

        Methods ``__getitem__`` and ``__contains__`` never fail or return False. Using a default with ``get`` will
        generate a value rather than using the provided default.

        >>> magic[1999]
        'Sofia'
        >>> magic[2019]
        '<Failed: id=2019>'
        >>> magic.get(2019, "foo")  # doctest: +SKIP
        '<Failed: id=2019>'

        **ID translation heuristics**

        Special handling for :py:class:`uuid.UUID` and ``UUID``-like strings improve matching.

        >>> string_uuid = "550e8400-e29b-41d4-a716-446655440000"
        >>> magic = MagicDict(
        ...     {string_uuid: "Found!"},
        ...     enable_uuid_heuristics=True,
        ... )
        >>> magic
        {UUID('550e8400-e29b-41d4-a716-446655440000'): 'Found!'}

        When ``enable_uuid_heuristics=True``, the ``MagicDict`` will attempt to cast keys to :class:`uuid.UUID`.

        >>> from uuid import UUID
        >>> magic[string_uuid], magic[UUID(string_uuid)]
        ('Found!', 'Found!')

        Keys that cannot be converted are left as-is.

        >>> magic["Hello"] = "World!"
        >>> magic["unknown"], magic["Hello"]
        ("<Failed: id='unknown'>", 'World!')

        To further customize ID matching behaviour, refer to the :class:`.Transformer` interface.
    """

    LOGGER = logging.getLogger(__package__).getChild("MagicDict")

    def __init__(
        self,
        real_translations: TranslatedIds[IdType],
        default_value: str = Format(Format.DEFAULT_FAILED).fstring(positional=True),
        enable_uuid_heuristics: bool = True,
        transformer: Transformer[IdType] | None = None,
    ) -> None:
        if enable_uuid_heuristics and real_translations:
            real_translations, enable_uuid_heuristics = _try_stringify_many(real_translations)
        if transformer:
            transformer.update_translations(real_translations)

        self._real: TranslatedIds[IdType] = real_translations
        self._default = self._verify_default_value(default_value)
        self._cast_key = enable_uuid_heuristics
        self._transformer = transformer

    def get(self, key: IdType, /, _: Any = None) -> str:
        """Same as ``__getitem__``.

        Values for missing keys are generated from :attr:`default_value`.
        """
        return self[key]

    def real_get(self, key: IdType) -> str | None:
        """Attempt to get an actual translation.

        This method behaves like ``MagicDict.__getitem__``, applying all appropriate heuristics **except** for falling
        back to the :attr:`default_value`. Returns ``None`` if the `key` cannot be mapped to a real values, like the
        regular ``dict.get`` method would.

        To bypass the heuristics, use :attr:`real` and :meth:`dict.get` instead. Note that the backing dict may still
        contain mappings added transformers, since the :meth:`.Transformer.update_translations` interface is called
        during initialization.
        """
        if key in self._real:
            return self._real[key]

        key = self._on_read(key)
        return self._real.get(key)

    def real_contains(self, key: IdType, /) -> bool:
        """Check if an actual translation exists using :meth:`real_get`."""
        return self.real_get(key) is not None

    @property
    def real(self) -> dict[IdType, str]:
        """Returns the backing dict."""
        return self._real

    @property
    def default_value(self) -> str:
        """Return the default string value to return for unknown keys, if any."""
        return self._default

    def _try_stringify(self, key: IdType) -> Any:
        return _uuid_utils.try_cast_one(key) if self._cast_key else key

    def __getitem__(self, key: IdType) -> str:
        value = self.real_get(key)
        return self.default_value.format(key) if value is None else value

    def __contains__(self, key: Any) -> bool:
        """Always returns ``True``."""
        return True  # We can always render something with default_value.

    def _on_read(self, key: IdType) -> IdType:
        key = self._try_stringify(key)
        if self._transformer and key not in self._real:
            self._transformer.try_add_missing_key(key, translations=self)
        return key

    def __setitem__(self, key: IdType, value: str) -> None:
        key = self._on_write(key)
        self._real[key] = value

    def __delitem__(self, key: IdType) -> None:
        key = self._on_write(key)
        del self._real[key]

    def _on_write(self, key: IdType) -> IdType:
        key = self._try_stringify(key)
        return key

    def __len__(self) -> int:
        return len(self._real)

    def __iter__(self) -> Iterator[IdType]:
        return iter(self._real)

    def __repr__(self) -> str:
        return repr(self._real)

    @classmethod
    def _verify_default_value(cls, default_value: str) -> str:
        for sample in "id", 0, UUID(int=0):
            try:
                default_value.format(sample)
                return default_value  # Formatting OK
            except KeyError as e:
                raise ValueError(f"Bad {default_value=}") from e
            except Exception:  # noqa: S110
                pass  # Attribute error, value error, etc. Happens naturally when the sample type is wrong.

        return default_value


def _try_stringify_many(real_translations: TranslatedIds[IdType]) -> tuple[TranslatedIds[IdType], bool]:
    original_ids = list(real_translations)
    if len(original_ids) == 0:
        return real_translations, True

    new_keys = _uuid_utils.try_cast_many(original_ids)
    if not isinstance(new_keys[0], UUID):
        return real_translations, False  # Keys are not UUID-like.

    uuid_translations = {uuid: real_translations[idx] for uuid, idx in zip(new_keys, original_ids, strict=True)}
    if len(real_translations) != len(uuid_translations):
        raise TypeError("Duplicate UUIDs found. Verify translation sources or set enable_uuid_heuristics=False.")

    return uuid_translations, True



"""Types used for offline translation."""

import dataclasses as _dataclasses
import typing as _t
from collections.abc import Sequence as _Sequence

from .. import types as _tt

if _t.TYPE_CHECKING:
    import pandas

    from ._format import Format

FormatType = _t.Union[str, "Format"]

PlaceholdersTuple = tuple[str, ...]
TranslatedIds = dict[_tt.IdType, str]  # {id: translation}

DictMakeTypes = _t.Mapping[str, _Sequence[_t.Any]] | _t.Mapping[_tt.IdType, str]


@_dataclasses.dataclass
class PlaceholderTranslations(_t.Generic[_tt.SourceType]):
    """Matrix of ID translation components returned by fetchers."""

    source: _tt.SourceType
    """Source from which translations were retrieved."""
    placeholders: PlaceholdersTuple
    """Names of placeholders in the order in which they appear in `records`."""
    records: _Sequence[_Sequence[_t.Any]]
    """Matrix of shape `N x M` where `N` is the number of IDs returned and `M` is the length of `placeholders`."""
    id_pos: int = -1
    """Position if the ID placeholder in `placeholders`."""
    placeholder_aliases: dict[str, str] = _dataclasses.field(default_factory=dict)
    """Alternative :attr:`placeholder <placeholders>` names."""

    @classmethod
    def make(cls, source: _tt.SourceType, data: "MakeTypes[_tt.SourceType, _tt.IdTypes]") -> _t.Self:
        """Try to make in instance from arbitrary input data.

        Args:
            source: Source label for the translations.
            data: Some data to convert to a ``PlaceholderTranslations`` instance.

        Returns:
            A new :class:`PlaceholderTranslations` instance.

        Raises:
            TypeError: If `data` cannot be converted.
        """
        try:
            from pandas import DataFrame  # noqa: PLC0415

            if isinstance(data, DataFrame):
                return cls.from_dataframe(source, data)
        except ImportError:
            pass

        if isinstance(data, dict):
            return cls.from_dict(source, data)
        if isinstance(data, PlaceholderTranslations):
            return cls(**_dataclasses.asdict(data))

        raise TypeError(data)  # pragma: no cover

    def to_dict(self, max_rows: int = 0) -> dict[str, _Sequence[_t.Any]]:
        """Create a dict representation of the translations."""
        records = self.records[:max_rows] if max_rows else self.records
        return {placeholder: [row[i] for row in records] for i, placeholder in enumerate(self.placeholders)}

    def to_pandas(self, max_rows: int = 0) -> "pandas.DataFrame":
        """Create a pandas DataFrame representation of the translations."""
        from pandas import DataFrame  # noqa: PLC0415

        return DataFrame.from_dict(self.to_dict(max_rows))

    @staticmethod
    def to_dicts(
        source_translations: "SourcePlaceholderTranslations[_tt.SourceType]",
        max_rows: int = 0,
    ) -> dict[_tt.SourceType, dict[str, _Sequence[_t.Any]]]:
        """Create a nested dict representation of the translations."""
        return {source: translations.to_dict(max_rows) for source, translations in source_translations.items()}

    @classmethod
    def from_dataframe(cls, source: _tt.SourceType, data: "pandas.DataFrame") -> _t.Self:
        """Create instance from a pandas DataFrame."""
        return cls(
            source,
            placeholders=tuple(data),
            records=data.to_numpy().tolist(),
            id_pos=data.columns.get_loc(_tt.ID) if _tt.ID in data else -1,
        )

    @classmethod
    def from_dict(cls, source: _tt.SourceType, data: DictMakeTypes[_tt.IdType]) -> _t.Self:
        """Create instance from a dict."""
        if cls._is_simple_form(data):
            # Users may pass dicts on the form {source: {id: name}}, where our data={id: name}.
            _t.assert_type(data, dict[_tt.IdType, str])
            records = tuple(tuple(item) for item in data.items())
            return cls(source, (_tt.ID, "name"), records, id_pos=0)

        try:
            from pandas import DataFrame  # noqa: PLC0415

            return cls.from_dataframe(source, DataFrame.from_dict(data))
        except ImportError:
            placeholders: tuple[str, ...] = tuple(data)  # type: ignore[arg-type]
            records = tuple(zip(*data.values(), strict=True))
            id_pos = placeholders.index(_tt.ID) if _tt.ID in placeholders else -1
            return cls(source, placeholders, records, id_pos)

    @classmethod
    def _is_simple_form(cls, data: DictMakeTypes[_tt.IdType]) -> _t.TypeGuard[dict[_tt.IdType, str]]:
        id_types = _t.get_args(_tt.IdTypes)
        return all(isinstance(key, id_types) and isinstance(value, str) for key, value in data.items())


MakeTypes = _t.Union[PlaceholderTranslations[_tt.SourceType], "pandas.DataFrame", DictMakeTypes[_tt.IdType]]
SourcePlaceholderTranslations = dict[_tt.SourceType, PlaceholderTranslations[_tt.SourceType]]



from abc import ABC, abstractmethod
from typing import Generic

from id_translation.fetching.types import FetchInstruction
from id_translation.offline.types import PlaceholderTranslations
from id_translation.types import IdType, SourceType

from ._fetcher import Fetcher


class CacheAccess(ABC, Generic[SourceType, IdType]):
    """Interface for user-managed caching.

    To enable caching, implement the abstract methods of the ``CacheAccess`` interface and pass it to the fetcher. See
    the :ref:`🚀 examples page <caching_example>` to get started.
    """

    def __init__(self) -> None:
        self._parent: Fetcher[SourceType, IdType] | None = None

    @property
    def enabled(self) -> bool:
        """Return the enabled status for this ``CacheAccess``.

        Returns ``True`` by default. If this property is ``False``, no other methods will be called.
        """
        return True

    @abstractmethod
    def load(
        self,
        instr: FetchInstruction[SourceType, IdType],
    ) -> PlaceholderTranslations[SourceType] | None:
        """Load cached translations.

        If this method returns ``None``, the ``AbstractFetcher`` will use :meth:`~.AbstractFetcher.fetch_translations`
        instead. The fetcher will then call :meth:`store` using `instr` and the newly fetched translations.

        Args:
            instr: A :class:`.FetchInstruction`.

        Returns:
            Cached :class:`.PlaceholderTranslations` or ``None``.
        """

    @abstractmethod
    def store(
        self,
        instr: FetchInstruction[SourceType, IdType],
        translations: PlaceholderTranslations[SourceType],
    ) -> None:
        """Store fetched translations.

        .. note::

           This method will never be called with translations that were returned by :meth:`load`.

        In other words, this method will only be called if ``CacheAccess.load(instr)`` returns ``None``.

        .. hint::

           The ``CacheAccess`` is under no obligation to actually store `translations`.

        For example, implementations may choose only to cache data when the :attr:`.FetchInstruction.fetch_all`-property
        of the given `instr` is ``True``.

        Args:
            instr: The :class:`.FetchInstruction` which produced the `translations`.
            translations: A :class:`.PlaceholderTranslations` produced by :meth:`~.AbstractFetcher.fetch_translations`.
        """

    @property
    def parent(self) -> Fetcher[SourceType, IdType]:
        """Parent :class:`.Fetcher` instance.

        The owner, typically an :class:`.AbstractFetcher`, should call :meth:`.set_parent` during initialization.

        Returns:
            The fetcher that owns this ``CacheAccess``.

        Raises:
            RuntimeError: If called before the parent is set.
        """
        if self._parent is None:
            raise RuntimeError("parent not set")
        return self._parent

    def set_parent(self, parent: Fetcher[SourceType, IdType]) -> None:
        """Set parent instance.

        Args:
            parent: A :class:`Fetcher`.

        Raises:
            RuntimeError: If a parent is already set.
        """
        if self._parent is not None:
            raise RuntimeError("parent already set")
        self._parent = parent



import logging
import warnings
from collections.abc import Iterable
from math import isinf
from time import perf_counter
from typing import Any, Generic, Self

from rics.collections.dicts import InheritedKeysDict
from rics.misc import get_by_full_name, tname
from rics.strings import format_perf_counter as fmt_perf
from rics.types import LiteralHelper

from .. import logging as _logging
from . import filter_functions as ff
from . import score_functions as sf
from ._cardinality import Cardinality
from ._directional_mapping import DirectionalMapping
from ._heuristic_score import HeuristicScore
from .exceptions import (
    BadFilterError,
    MappingError,
    UnmappedValuesError,
    UnmappedValuesWarning,
    UserMappingError,
    UserMappingWarning,
)
from .types import (
    CandidateType,
    CardinalityType,
    ContextType,
    FilterFunction,
    OnUnknownUserOverride,
    OnUnmapped,
    ScoreFunction,
    UserOverrideFunction,
    ValueType,
)

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=UserWarning)
    from .matrix import ScoreHelper, ScoreMatrix

inf = float("inf")
FilterFunctionArgItem = tuple[str | FilterFunction[ValueType, CandidateType, ContextType], dict[str, Any]]
FilterFunctionWithKwargs = tuple[FilterFunction[ValueType, CandidateType, ContextType], dict[str, Any]]


class Mapper(Generic[ValueType, CandidateType, ContextType]):  # noqa: PLW1641
    """Optimal value-candidate matching.

    For an introduction to mapping, see the :ref:`mapping-primer` page.

    Args:
        score_function: A callable which accepts a value `k` and an ordered collection of candidates `c`, returning a
            score ``s_i`` for each candidate `c_i` in `c`. Default: ``s_i = float(k == c_i)``. Higher=better match.
        score_function_kwargs: Keyword arguments for `score_function`.
        filter_functions: Function-kwargs pairs of filters to apply before scoring.
        min_score: Minimum score `s_i`, as given by ``score(k, c_i)``, to consider `k` a match for `c_i`.
        overrides: If a dict, assumed to be 1:1 mappings (`value` to `candidate`) which override the scoring logic. If
            :class:`rics.collections.dicts.InheritedKeysDict`, the context passed to :meth:`apply` is used to retrieve
            specific overrides.
        on_unmapped: Action to take if mapping fails for any values.
        on_unknown_user_override: Action to take if an :attr:`~id_translation.mapping.types.UserOverrideFunction`
            returns an unknown candidate. Unknown candidates, i.e. candidates not in the input `candidates` collection,
            will not be used unless `'allow'` is chosen.
        cardinality: Desired cardinality for mapped values. Derive for each matching if ``None``.
    """

    def __init__(
        self,
        score_function: str | ScoreFunction[ValueType, CandidateType, ContextType] = "disabled",
        score_function_kwargs: dict[str, Any] | None = None,
        filter_functions: Iterable[FilterFunctionArgItem[ValueType, CandidateType, ContextType]] = (),
        min_score: float = 0.90,
        overrides: dict[ValueType, CandidateType]
        | InheritedKeysDict[ContextType, ValueType, CandidateType]
        | None = None,
        on_unmapped: OnUnmapped = "ignore",
        on_unknown_user_override: OnUnknownUserOverride = "raise",
        cardinality: CardinalityType | None = Cardinality.ManyToOne,
    ) -> None:
        if min_score <= 0 or isinf(min_score):
            raise ValueError(f"Got {min_score=}. The score limit should be a finite positive value.")

        self._score = get_by_full_name(score_function, sf) if isinstance(score_function, str) else score_function
        self._score_kwargs = score_function_kwargs or {}
        self._min_score = min_score
        self._overrides: InheritedKeysDict[ContextType, ValueType, CandidateType] | dict[ValueType, CandidateType] = (
            overrides if isinstance(overrides, InheritedKeysDict) else (overrides or {})
        )
        self._on_unmapped = OU_HELPER.check(on_unmapped)
        self._on_unknown_user_override = OUUO_HELPER.check(on_unknown_user_override)
        self._cardinality = None if cardinality is None else Cardinality.parse(cardinality, strict=True)

        self._filters = self._initialize_filter_functions(filter_functions)
        self._logger = logging.getLogger(__package__).getChild("Mapper")  # This will almost always be overwritten

    def apply(
        self,
        values: Iterable[ValueType],
        candidates: Iterable[CandidateType],
        context: ContextType | None = None,
        override_function: UserOverrideFunction[ValueType, CandidateType, ContextType] | None = None,
        *,
        task_id: int | None = None,
        **kwargs: Any,
    ) -> DirectionalMapping[ValueType, CandidateType]:
        """Map values to candidates.

        Args:
            values: Iterable of elements to match to candidates.
            candidates: Iterable of candidates to match with `value`. Duplicate elements will be discarded.
            context: Context in which mapping is being done.
            override_function: A callable that takes inputs ``(value, candidates, context)`` that returns either
                ``None`` (let the regular mapping logic decide) or one of the `candidates`. How non-candidates returned
                is handled is determined by the :attr:`on_unknown_user_override` property.
            task_id: Used for logging.
            **kwargs: Runtime keyword arguments for score and filter functions. May be used to add information which is
                not known when the ``Mapper`` is initialized.

        Returns:
            A :class:`.DirectionalMapping` on the form ``{value: [matched_candidates..]}``. May be turned into a
            plain dict ``{value: candidate}`` by using the :meth:`.DirectionalMapping.flatten` function (only if
            :attr:`.DirectionalMapping.cardinality` is of type :attr:`.Cardinality.one_right`).

        Raises:
            MappingError: If any values failed to match and ``on_unmapped='raise'``.
            BadFilterError: If a filter returns candidates that are not a subset of the original candidates.
            UserMappingError: If `override_function` returns an unknown candidate and
                ``on_unknown_user_override != 'allow'``
            MappingError: If passing ``context=None`` (the default) when using context-sensitive overrides (type
                :class:`rics.collections.dicts.InheritedKeysDict`).
        """
        start = perf_counter()

        logger = self.logger

        candidates = list(candidates)
        values = list(values)
        if not (values and candidates):
            logger.debug(
                "Aborting since values=%r and candidates=%r in context=%r.",
                values,
                candidates,
                context,
                extra={"task_id": task_id},
            )
            return DirectionalMapping(left_to_right={}, _verify=False, cardinality=self.cardinality)

        scores = self.compute_scores(values, candidates, context, override_function, task_id=task_id, **kwargs)

        dm: DirectionalMapping[ValueType, CandidateType] = self.to_directional_mapping(scores, task_id=task_id)

        unmapped = scores.get_finite_values().difference(dm.left)
        if unmapped:
            extra = f" in {context=}" if context else ""
            candidates = set(scores.candidates)  # Includes candidates added by override logic.
            self._report_unmapped(f"Could not map {unmapped}{extra} to any of {candidates=}.", task_id=task_id)

        logger = self.logger
        if _logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG):
            cardinality = "automatic" if self.cardinality is None else self.cardinality.name

            l2r = dm.left_to_right
            matches = " Matches:\n" + "\n".join(
                f"    {v!r} -> {repr(l2r[v]) if v in l2r else '<no matches>'}" for v in values
            )

            logger.debug(
                f"Mapping with {cardinality=} completed for {values}x{candidates} in {fmt_perf(start)}."
                f"{matches}\nMatched {len(dm.left)}/{len(values)} values with {len(dm.right)} different candidates.",
                extra={"task_id": task_id},
            )

        return dm

    def _report_unmapped(self, msg: str, task_id: int | None) -> None:
        logger = self.logger

        if self.on_unmapped == "raise":
            msg += "\nHint: Set on_unmapped='warn' or on_unmapped='ignore' to allow unmapped values."
            logger.error(msg, extra={"task_id": task_id})
            raise UnmappedValuesError(msg)
        elif self.on_unmapped == "warn":
            logger.warning(msg, extra={"task_id": task_id})
            msg += (
                "\nHint: Set on_unmapped='ignore' to hide this warning, or "
                f"on_unmapped='raise' to raise an {UnmappedValuesError.__name__}."
            )
            warnings.warn(msg, UnmappedValuesWarning, stacklevel=3)
        elif _logging.ENABLE_VERBOSE_LOGGING:
            logger.debug(msg, extra={"task_id": task_id})

    def compute_scores(
        self,
        values: Iterable[ValueType],
        candidates: Iterable[CandidateType],
        context: ContextType | None = None,
        override_function: UserOverrideFunction[ValueType, CandidateType, ContextType] | None = None,
        task_id: int | None = None,
        **kwargs: Any,
    ) -> ScoreMatrix[ValueType, CandidateType]:
        """Compute likeness scores.

        Args:
            values: Iterable of elements to match to candidates.
            candidates: Iterable of candidates to match with `value`. Duplicate elements will be discarded.
            context: Context in which mapping is being done.
            override_function: A callable that takes inputs ``(value, candidates, context)`` that returns either
                ``None`` (let the regular mapping logic decide) or one of the `candidates`. How non-candidates returned
                is handled is determined by the :attr:`on_unknown_user_override` property.
            task_id: Used for logging.
            **kwargs: Runtime keyword arguments for score and filter functions. May be used to add information which is
                not known when the ``Mapper`` is initialized.

        Returns:
            A ``DataFrame`` of value-candidate match scores, with ``DataFrame.index=values`` and
            ``DataFrame.columns=candidates``.

        Raises:
            BadFilterError: If a filter returns candidates that are not a subset of the original candidates.
            UserMappingError: If `override_function` returns an unknown candidate and
                ``on_unknown_user_override != 'allow'``
        """
        start = perf_counter()

        scores = ScoreMatrix(values, candidates)
        values = scores.values
        candidates = scores.candidates

        extra = f" in {context=}" if context else ""

        logger = self.logger
        logger_enabled = _logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG)
        if scores.size == 0:
            if logger_enabled:
                end = "" if (values or candidates) else ", but got neither"
                logger.warning(
                    f"Abort mapping{extra} of {values}x{candidates}. Both values and candidates must be given{end}."
                )
            return scores

        score_fn = self._score
        if isinstance(score_fn, HeuristicScore):
            self._score_kwargs["task_id"] = task_id

        if logger_enabled:
            logger.debug(
                f"Begin computing match scores{extra} for {values}x{candidates} using {score_fn}.",
                extra={"task_id": task_id},
            )

        unmapped_values = self._handle_overrides(scores, context, override_function, task_id)

        logger_enabled = _logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG)
        for value in unmapped_values:
            filtered_candidates = self._apply_filters(value, candidates, context, kwargs, task_id)
            if not filtered_candidates:
                continue

            scores_for_value: Iterable[float]
            if value in filtered_candidates:
                scores_for_value = [(inf if value == c else -inf) for c in filtered_candidates]  # Identity match
            else:
                if logger_enabled:
                    logger.debug(f"Compute match scores for {value=}.", extra={"task_id": task_id})
                scores_for_value = score_fn(value, filtered_candidates, context, **self._score_kwargs, **kwargs)

            for score, candidate in zip(scores_for_value, filtered_candidates, strict=True):
                scores[value, candidate] = score

        if (len(scores.values) > 1 or _logging.ENABLE_VERBOSE_LOGGING) and logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Computed {len(scores.values)}x{len(scores.candidates)} "
                f"match scores in {context=} in {fmt_perf(start)}:\n{scores.to_string()}",
                extra={"task_id": task_id},
            )
        return scores

    def to_directional_mapping(
        self,
        scores: ScoreMatrix[ValueType, CandidateType],
        *,
        task_id: int | None = None,
    ) -> DirectionalMapping[ValueType, CandidateType]:
        """Create a ``DirectionalMapping`` from match scores.

        Args:
            scores: A score matrix, where ``scores.index`` are values and ``score.columns`` are treated as the
                candidates.
            task_id: Used for logging.

        Returns:
            A ``DirectionalMapping``.

        See Also:
            :meth:`.ScoreHelper.to_directional_mapping`
        """
        helper = ScoreHelper(scores, self._min_score, self._logger, task_id=task_id)
        return helper.to_directional_mapping(self.cardinality)

    @property
    def cardinality(self) -> Cardinality | None:
        """Return upper cardinality bound during mapping."""
        return self._cardinality

    @property
    def on_unmapped(self) -> OnUnmapped:
        """Return the action to take if mapping fails for any values."""
        return self._on_unmapped

    @property
    def on_unknown_user_override(self) -> OnUnknownUserOverride:
        """Return the action to take if an override function returns an unknown candidate.

        Returns:
            Action to take if a user-defined override function returns an unknown candidate.
        """
        return self._on_unknown_user_override

    @property
    def logger(self) -> logging.Logger:
        """Return the ``Logger`` that is used by this instance."""
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger) -> None:
        self._logger = logger

    def _handle_overrides(
        self,
        scores: ScoreMatrix[ValueType, CandidateType],
        context: ContextType | None,
        override_function: UserOverrideFunction[ValueType, CandidateType, ContextType] | None,
        task_id: int | None,
    ) -> list[ValueType]:
        applied: dict[ValueType, CandidateType] = {}

        def apply(v: ValueType, oc: CandidateType) -> None:
            scores[v, :] = -inf
            scores[v, oc] = inf
            unmapped_values.remove(v)
            applied[v] = oc

        unmapped_values = scores.values

        logger = self.logger
        logger_enabled = _logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG)
        if override_function:
            for value, override_candidate in self._get_function_overrides(
                override_function,
                scores.values,
                scores.candidates,
                context,
                task_id,
            ):
                if logger_enabled:
                    logger.debug(
                        f"Using override {value!r} -> {override_candidate!r} returned by {override_function=}.",
                        extra={"task_id": task_id},
                    )
                apply(value, override_candidate)

        for value, override_candidate in self._get_static_overrides(unmapped_values, context).items():
            apply(value, override_candidate)

        if logger_enabled and (self._overrides or override_function is not None):
            num_overrides = len(self._overrides) + int(override_function is not None)
            result = f"and found {len(applied)} matches={applied} in" if applied else "but none were a match for"
            done = "All values mapped by overrides. " if (not unmapped_values and applied) else ""
            logger.debug(
                f"{done}Applied {num_overrides} overrides, {result} the given values={scores.values}.",
                extra={"task_id": task_id},
            )

        return unmapped_values

    def _get_static_overrides(
        self,
        values: Iterable[ValueType],
        context: ContextType | None,
    ) -> dict[ValueType, CandidateType]:
        if not self._overrides:
            return {}

        if isinstance(self._overrides, InheritedKeysDict):
            if context is None:
                raise MappingError("Must pass a context in context-sensitive mode.")
            overrides = self._overrides.get(context, {})
        else:
            overrides = self._overrides

        return {value: overrides[value] for value in overrides if value in values}

    def _get_function_overrides(
        self,
        func: UserOverrideFunction[ValueType, CandidateType, ContextType],
        values: Iterable[ValueType],
        candidates: Iterable[CandidateType],
        context: ContextType | None,
        task_id: int | None,
    ) -> list[tuple[ValueType, CandidateType]]:
        candidates = set(candidates)

        logger = self.logger

        ans = []
        for value in values:
            user_override = func(value, candidates, context)
            if user_override is None:
                continue

            if user_override not in candidates and self.on_unknown_user_override != "keep":
                msg = f"The user-defined override function {func} returned an unknown candidate={user_override!r} for {value=}."
                note = "Hint: Set on_unknown_user_override='keep' to use this value anyway."
                if self.on_unknown_user_override == "raise":
                    logger.error(msg, extra={"task_id": task_id})
                    exc = UserMappingError(msg, value, candidates)
                    exc.add_note(note)
                    raise exc

                assert self.on_unknown_user_override == "warn", f"bad {self.on_unknown_user_override=}"  # noqa: S101
                logger.warning(msg, extra={"task_id": task_id})

                msg += f"\n{note}"
                warnings.warn(msg, UserMappingWarning, stacklevel=2)

                continue

            ans.append((value, user_override))

        return ans

    def _apply_filters(
        self,
        value: ValueType,
        candidates: Iterable[CandidateType],
        context: ContextType | None,
        kwargs: dict[str, Any],
        task_id: int | None,
    ) -> set[CandidateType]:
        candidates = list(candidates)
        filtered_candidates = set(candidates)

        for filter_function, function_kwargs in self._filters:
            if "task_id" in function_kwargs:
                function_kwargs["task_id"] = task_id
            filtered_candidates = filter_function(value, filtered_candidates, context, **function_kwargs, **kwargs)

            not_in_original_candidates = filtered_candidates.difference(candidates)
            if not_in_original_candidates:
                raise BadFilterError(
                    f"Filter {tname(filter_function)}({value}, candidates, **{kwargs}) created new"
                    f"candidates: {not_in_original_candidates}"
                )

            if not filtered_candidates:
                break

        logger = self.logger
        logger_enabled = _logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG)
        if logger_enabled and len(self._filters):
            diff = set(candidates).difference(filtered_candidates)
            removed = f"removing candidates={diff}" if diff else "but did not remove any candidates"
            done = "All candidates removed by filtering. " if not filtered_candidates else ""
            logger.debug(
                f"{done}Applied {len(self._filters)} filters for {value=}, {removed}.",
                extra={"task_id": task_id},
            )

        return filtered_candidates

    def __repr__(self) -> str:
        score = self._score
        return f"{tname(self)}({score=} >= {self._min_score}, {len(self._filters)} filters)"

    def copy(self, **overrides: Any) -> Self:
        """Make a copy of this ``Mapper``.

        Args:
            overrides: Keyword arguments to use when instantiating the copy. Options that aren't given will be taken
                from the current instance. See the :class:`Mapper` class documentation for possible choices.

        Returns:
            A copy of this ``Mapper`` with `overrides` applied.
        """
        kwargs: dict[str, Any] = {
            "score_function": self._score,
            "min_score": self._min_score,
            "on_unmapped": self.on_unmapped,
            "on_unknown_user_override": self.on_unknown_user_override,
            "cardinality": self.cardinality,
            **overrides,
        }

        if "score_function_kwargs" not in kwargs:
            kwargs["score_function_kwargs"] = self._score_kwargs.copy()

        if "filter_functions" not in kwargs:
            kwargs["filter_functions"] = [(func, func_kwargs.copy()) for func, func_kwargs in self._filters]

        if "overrides" not in kwargs:
            kwargs["overrides"] = self._overrides.copy()

        cls = type(self)
        return cls(**kwargs)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Mapper):
            return False

        return all(
            (
                self._score == other._score,
                self._score_kwargs == other._score_kwargs,
                self._filters == other._filters,
                self._min_score == other._min_score,
                self._overrides == other._overrides,
                self._on_unmapped == other._on_unmapped,
                self._on_unknown_user_override == other._on_unknown_user_override,
                self._cardinality == other._cardinality,
            )
        )

    @classmethod
    def _initialize_filter_functions(
        cls,
        filter_functions: Iterable[FilterFunctionArgItem[ValueType, CandidateType, ContextType]],
    ) -> list[FilterFunctionWithKwargs[ValueType, CandidateType, ContextType]]:
        rv = []
        for func_or_str, kwargs in filter_functions:
            func = get_by_full_name(func_or_str, ff) if isinstance(func_or_str, str) else func_or_str

            if "task_id" in getattr(func, "__annotations__", {}):
                kwargs.setdefault("task_id", None)

            rv.append((func, kwargs))
        return rv


OU_HELPER: LiteralHelper[OnUnmapped] = LiteralHelper(
    OnUnmapped,
    default_name="on_unmapped",
    type_name="OnUnmapped",
)
OUUO_HELPER: LiteralHelper[OnUnknownUserOverride] = LiteralHelper(
    OnUnknownUserOverride,
    default_name="on_unknown_user_override",
    type_name="OnUnknownUserOverride",
)



from collections.abc import Iterable, Mapping
from typing import Any, Literal, Self

from rics.misc import tname

from . import parse_format_string
from .types import PlaceholdersTuple


class Format:
    """Format specification for translations strings.

    Translation formats are similar to regular f-strings, with two important exceptions:

    1. Positional placeholders (``'{}'``) may not be used; correct form is ``'{placeholder-name}'``.
    2. Placeholders surrounded by ``'[]'`` denote an optional element. Optional elements are rendered...

       * Only if `all` of its placeholders are defined.
       * Without delimiting brackets.
       * As literal text (with brackets) if there is no placeholder in the block.

    .. hint::

       Double the wanted bracket character to render as a literal, analogous to ``'{{'`` and ``'}}'``
       in plain Python f-strings. See the example below for a demonstration.

    Args:
        fmt: A translation fstring.

    Examples:
        **Basic usage**

        Formats are created by passing a single ``str`` arguments, as described above.

        >>> Format(Format.DEFAULT)
        "Format('{id}:{name}')"
        >>> Format(Format.DEFAULT).fstring().format(id=1, name="First")
        '1:First'

        Using :meth:`Format.fstring` and :py:meth:`str.format` is flexible but verbose. Formats can be applied either
        through :meth:`Format.format`...

        >>> fmt = Format(Format.DEFAULT_FAILED)
        >>> fmt.format(id=1, name="First")

        ...or just ``Format.__call__()``.

        >>> fmt(id=1, name="First")
        '<Failed: id=1>'

        Using either convenience method will use as many placeholders as possible.

        >>> fmt = Format(Format.DEFAULT)
        >>> fmt.placeholders
        "('id', 'name')"
        >>> fmt(id=1, name="First")
        '1:First'
        >>> fmt(id=1, name="First", unknown=20.19)
        '1:First'

        Unknown placeholders are simply ignored.

        **Optional placeholders**

        A format string using literal angle brackets and an optional element.

        >>> from id_translation.offline import Format
        >>> fmt = Format("{id}:[[{name}]][, nice={is_nice}]")

        The ``Format`` class when used directly only returns required placeholders by default...

        >>> fmt.fstring()
        '{id}:[{name}]'
        >>> fmt(id=0, name="Tarzan")
        '0:[Tarzan]'

        ...but the :attr:`placeholders` attribute can be used to retrieve all placeholders, required and optional:

        >>> fmt.placeholders
        ('id', 'name', 'is_nice')
        >>> fmt(id=1, name="Morris", is_nice=True)
        '1:[Morris], nice=True'

    The :class:`.Translator` will automatically add optional placeholders, if they are present in the source.

    .. note::
       Python format specifications and conversions are preserved.

    This is especially useful for long values such as UUIDs.

    >>> from uuid import UUID
    >>> uuid = UUID("550e8400-e29b-41d4-a716-446655440000")

    Convert to string and truncate to eight characters.

    >>> Format("{id!s:.8}:{name!r}").format(id=uuid, name="Sofia")
    "550e8400:'Sofia'"

    See the official :py:ref:`formatspec` documentation for details.
    """

    DEFAULT: Literal["{id}:{name}"] = "{id}:{name}"
    """Default translation format."""

    DEFAULT_FAILED: Literal["<Failed: id={id!r}>"] = "<Failed: id={id!r}>"
    """Default format for missing IDs."""

    def __init__(self, fmt: str) -> None:
        self._fmt = fmt
        self._elements: list[parse_format_string.Element] = parse_format_string.get_elements(fmt)

    def format(self, **placeholders: Any) -> str:
        """Apply the format.

        Args:
            **placeholders: Formats to use in the finals string.

        Returns:
            Formatting using `placeholders`.
        """
        return self.fstring(placeholders, positional=False).format_map(placeholders)

    __call__ = format

    def fstring(self, placeholders: Iterable[str] | None = None, *, positional: bool = False) -> str:
        """Create a format string for the given placeholders.

        Args:
            placeholders: Keys to keep. Passing ``None`` is equivalent to passing :attr:`required_placeholders`.
            positional: If ``True``, remove names to return a positional fstring.

        Returns:
            An fstring with optional elements removed unless included in `placeholders`.

        Raises:
            KeyError: If required placeholders are missing.
        """
        placeholders = placeholders or self.required_placeholders
        missing_required_placeholders = set(self.required_placeholders).difference(placeholders)
        if missing_required_placeholders:
            raise KeyError(f"Required key(s) {missing_required_placeholders} missing from {placeholders=}.")

        return self._make_fstring(placeholders, positional=positional)

    def _make_fstring(self, placeholders: Iterable[str], positional: bool) -> str:
        def predicate(e: parse_format_string.Element) -> bool:
            return e.required or set(placeholders).issuperset(e.placeholders)

        return "".join(e.positional_part if positional else e.part for e in filter(predicate, self._elements))

    def partial(self, defaults: Mapping[str, Any]) -> Self:
        """Get a partially formatted :meth:`fstring`.

        Args:
            defaults: Keys which should be replaced with real values. Keys which are **not** part of `defaults` will
                be left as-is.

        Returns:
            A partially formatted fstring.
        """
        new_fmt, _placeholders = parse_format_string.Element.parse_block(self._fmt, defaults=defaults)
        cls = type(self)
        return cls(new_fmt)

    @classmethod
    def parse(cls, fmt: str | Self) -> Self:
        """Parse a format.

        Args:
            fmt: A ``str`` or ``Format`` instance.

        Returns:
            A ``Format`` instance.
        """
        return cls(fmt) if isinstance(fmt, str) else fmt

    @property
    def placeholders(self) -> PlaceholdersTuple:
        """All placeholders in the order in which they appear."""
        return self._extract_placeholders(self._elements)

    @property
    def required_placeholders(self) -> PlaceholdersTuple:
        """All required placeholders in the order in which they appear."""
        return self._extract_placeholders(filter(lambda e: e.required, self._elements))

    @property
    def optional_placeholders(self) -> PlaceholdersTuple:  # pragma: no cover
        """All optional placeholders in the order in which they appear."""
        return self._extract_placeholders(filter(lambda e: not e.required, self._elements))

    @staticmethod
    def _extract_placeholders(elements: Iterable[parse_format_string.Element]) -> PlaceholdersTuple:
        ans = []
        for e in elements:
            ans.extend(e.placeholders)
        return tuple(ans)

    def __repr__(self) -> str:
        return f"{tname(self)}({self._fmt!r})"



"""Types used for mapping."""

import typing as _t
from collections import abc as _abc

if _t.TYPE_CHECKING:
    from ._cardinality import Cardinality

HL = _t.TypeVar("HL", bound=_abc.Hashable)
"""Hashable type on the left side of a directional relationship."""
HR = _t.TypeVar("HR", bound=_abc.Hashable)
"""Hashable type on the right side of a directional relationship."""
LeftToRight = dict[HL, tuple[HR, ...]]
"""A left-to-right mapping."""
RightToLeft = dict[HR, tuple[HL, ...]]
"""A right-to-left mapping."""

ValueType = _t.TypeVar("ValueType", bound=_abc.Hashable)
"""A type of item being mapped."""
CandidateType = _t.TypeVar("CandidateType", bound=_abc.Hashable)
"""A type of item being mapped."""
MatchTuple = tuple[CandidateType, ...]
"""A tuple of candidates matched to a value."""
ContextType = _t.TypeVar("ContextType", bound=_abc.Hashable)
"""Type of context in which mapping is being performed."""

CardinalityType = _t.Union[str, "Cardinality"]

UserOverrideFunction = _abc.Callable[
    [ValueType, set[CandidateType], ContextType | None],
    CandidateType | None,
]
"""Signature for a user-defined override function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which scoring is being performed.

Returns:
    Either ``None`` (let regular logic decide) or a single candidate `c` in `candidates`.
"""

ScoreFunction = _abc.Callable[
    [ValueType, _abc.Iterable[CandidateType], ContextType | None],
    _abc.Iterable[float],
]
"""Signature for a likeness score function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which scoring is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Yields:
    A score for each candidate `c` in `candidates`.
"""

AliasFunction = _abc.Callable[
    [ValueType, _abc.Iterable[CandidateType], ContextType | None],
    tuple[ValueType, _abc.Iterable[CandidateType]],
]
"""Signature for an alias function for heuristic scoring.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which mapping is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Returns:
    A tuple (name, candidates) with applied heuristics to increase (or decrease) score as desired.
"""

FilterFunction = _abc.Callable[
    [ValueType, _abc.Iterable[CandidateType], ContextType | None],
    set[CandidateType],
]
"""Signature for a filter function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which filtering is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Returns:
    A subset of candidates to keep.
"""

HeuristicsTypes: _t.TypeAlias = (
    AliasFunction[ValueType, CandidateType, ContextType] | FilterFunction[ValueType, CandidateType, ContextType]
)
"""Types that may be interpreted as a score function heuristic."""

OnUnmapped = _t.Literal["raise", "warn", "ignore"]
"""Action types for unmapped values."""
OnUnknownUserOverride = _t.Literal["raise", "warn", "keep"]
"""Action types for bad user overrides (dict or function)."""



from time import perf_counter
from typing import TYPE_CHECKING, Generic

from rics.misc import get_public_module

from ..logging import generate_task_id

if TYPE_CHECKING:
    from .._translator import Translator

from ..dio import DataStructureIO, resolve_io
from ..types import IdType, NameType, SourceType, Translatable


class BaseTask(Generic[NameType, SourceType, IdType]):
    """Internal type; not part of the public API."""

    def __init__(
        self,
        caller: "Translator[NameType, SourceType, IdType]",
        translatable: Translatable[NameType, IdType],
        task_id: int | None = None,
    ) -> None:
        self.caller = caller

        self.translatable = translatable

        self._io: DataStructureIO[Translatable[NameType, IdType], NameType, SourceType, IdType]
        self._start = perf_counter()
        self._task_id = generate_task_id(self._start) if task_id is None else task_id

        self._io = resolve_io(translatable, task_id=self.task_id)
        self._type_name: str | None = None
        self._full_type_name: str | None = None

        self._timings: dict[str, float] = {}

        if caller.online:
            caller.fetcher.initialize_sources(self.task_id)

    def add_timing(self, key: str, value: float, /) -> None:
        """Add timing."""
        # assert key not in self._timings, f"duplicate {key=}"
        self._timings[key] = value

    def get_timings_ms(self) -> dict[str, float]:
        """Retrieve timings."""
        return {k: round(1000 * v, 3) for k, v in self._timings.items()}

    @property
    def io(self) -> DataStructureIO[Translatable[NameType, IdType], NameType, SourceType, IdType]:
        """Initialized :class:`DataStructureIO` instance."""
        return self._io

    @property
    def type_name(self) -> str:
        """Stylized type name, e.g. `'DataFrame'`."""
        if self._type_name is None:
            self._type_name = repr(type(self.translatable).__name__)
        return self._type_name

    @property
    def full_type_name(self) -> str:
        """Canonical type name, e.g. `pandas.DataFrame`."""
        if self._full_type_name is None:
            self._full_type_name = get_public_module(type(self.translatable), resolve_reexport=True, include_name=True)
        return self._full_type_name

    @property
    def task_id(self) -> int:
        """Generated ID for this task. Used for logging."""
        return self._task_id



"""Functions that return a subset of candidates with which to continue the matching procedure.

Mapping of the current `value` is aborted if an empty set is returned. Functions such as :func:`filter_names` and
:func:`filter_sources` use this to allow (or disallow) names and sources that match a given regex pattern.
"""

import logging
import re
from collections.abc import Iterable as _Iterable
from typing import Any as _Any

from .. import logging as _logging
from ..types import ID
from . import exceptions


def filter_names(
    value: str,
    candidates: _Iterable[str],
    context: _Any,
    regex: str,
    remove: bool = False,
    *,
    task_id: int | None = None,
) -> set[str]:
    """Filter names to translate based on `regex`.

    Analogous to the built-in :py:func:`filter`-function, ``filter_names`` keeps only the names (`value`) that match the
    given `regex`. This behavior may be reversed by setting the `remove` flag to ``True``.

    Args:
        value: A name that should be mapped one of the sources in `candidates`.
        candidates: Candidate sources.
        context: Should be ``None``. Always ignored, exists for compatibility.
        regex: A regex pattern. Will be matched against the `value`.
        remove: If ``True``, remove matching values.
        task_id: Used for logging.

    Returns:
        The original candidates if `value` matches the given `regex`. An empty set, otherwise.

    Examples:
        Ensuring that untranslatable IDs are left as-is.

        >>> sources = {"employees", "countries", "orders"}
        >>> name = "employee_id"
        >>> allowed = filter_names(
        ...     name,
        ...     candidates=sources,
        ...     context=None,
        ...     regex=".*_id$",
        ... )
        >>> sorted(allowed)
        ['countries', 'employees', 'orders']

        The call above kept the `'employee_id'` name (by returning all candidate sources).
    """
    function_name = filter_names.__name__
    _check_context(function_name, context=context, want_none=True)
    keep = _filter_single(
        value,
        regex=regex,
        remove=remove,
        label="name",
        function_name=function_name,
        action="Do not translate",
        task_id=task_id,
    )
    return set(candidates) if keep else set()


def filter_sources(
    value: str,
    candidates: _Iterable[str],
    context: _Any,
    regex: str,
    remove: bool = False,
    *,
    task_id: int | None = None,
) -> set[str]:
    """Filter sources based on `regex`.

    Analogous to the built-in :py:func:`filter`-function, ``filter_sources`` keeps only the sources (`context`) that
    match the given `regex`. This behavior may be reversed by setting the `remove` flag to ``True``.

    Args:
        value: Target placeholder.
        candidates: Available placeholders in the source named by `context`. Always ignored, exists for compatibility.
        context: The source to which the `candidates` belong.
        regex: A regex pattern. Will be matched against the `context`.
        remove: If ``True``, remove matching values.
        task_id: Used for logging.

    Returns:
        The original candidates if `context` matches the given `regex`. An empty set, otherwise.

    Examples:
        Avoiding uninteresting sources (for ID translation purposes).

        >>> source = "some_metadata_table"
        >>> allowed = filter_sources(
        ...     "id",
        ...     candidates={"id", "name", "some_other_column"},
        ...     context=source,
        ...     regex=".*metadata.*",
        ...     remove=True,
        ... )
        >>> len(allowed)
        0

        The call above filtered out the `'some_metadata_table'` source (by removing all candidates).

    Notes:
        Returns immediately if `value` != `'id'`, to avoid unnecessary work. The
    """
    if value != ID:
        return set(candidates)

    function_name = filter_names.__name__
    _check_context(function_name, context=context, want_none=False)
    keep = _filter_single(
        context,
        regex=regex,
        remove=remove,
        label="source",
        function_name=function_name,
        task_id=task_id,
    )
    return set(candidates) if keep else set()


def filter_placeholders(
    value: str,  # noqa: ARG001
    candidates: _Iterable[str],
    context: _Any,
    regex: str,
    remove: bool = False,
    task_id: int | None = None,
) -> set[str]:
    """Filter placeholders, as they appear in the source given by `context`, based on `regex`.

    Args:
        value: Target placeholder. Always ignored, exists for compatibility.
        candidates: Available placeholders in the source named by `context`.
        context: The source to which the `candidates` belong.
        regex: A regex pattern. Will be matched against elements of the `candidates`.
        remove: If ``True``, remove matching values.
        task_id: Used for logging.

    Returns:
        Placeholders that may be used.

    Examples:
        Removing irrelevant but possibly confusing columns.

        >>> actual_placeholders = {"id", "name", "old_id", "previous_id"}
        >>> allowed = filter_placeholders(
        ...     value="ignored",
        ...     candidates=actual_placeholders,
        ...     context="ignored",
        ...     regex="^(old|previous).*",
        ...     remove=True,
        ... )
        >>> sorted(allowed)
        ['id', 'name']
    """
    _check_context(filter_placeholders.__name__, context=context, want_none=False)

    pattern = re.compile(regex, flags=re.IGNORECASE)
    candidates = set(candidates)
    ans = {c for c in candidates if (pattern.match(c) is None) is remove}

    if _logging.ENABLE_VERBOSE_LOGGING and len(ans) < len(candidates):
        logger = logging.getLogger(__name__).getChild(filter_placeholders.__name__)
        if logger.isEnabledFor(logging.DEBUG):
            removed = candidates.difference(ans)
            logger.debug(
                f"Discard placeholders={removed!r} in source={context!r}; matches {pattern=}.",
                extra={"task_id": task_id},
            )

    return ans


def _check_context(name: str, context: _Any, want_none: bool) -> None:
    if (context is None) is want_none:
        return

    purpose = "name-to-source" if want_none else "placeholder"
    raise exceptions.BadFilterError(f"Function '{__name__}.{name}' should only be used for {purpose} mapping.")


def _filter_single(
    string: str,
    regex: str,
    remove: bool,
    label: str,
    function_name: str,
    action: str = "Discard",
    *,
    task_id: int | None = None,
) -> bool:
    keep = (re.match(regex, string, re.IGNORECASE) is None) is remove

    if keep or not _logging.ENABLE_VERBOSE_LOGGING:
        return keep

    logger = logging.getLogger(__name__).getChild(function_name)
    if logger.isEnabledFor(logging.DEBUG):
        pattern = re.compile(regex, re.IGNORECASE)
        logger.debug(
            f"{action} {label}={string!r}; {'matches' if remove else 'does not match'} {pattern=}.",
            extra={"task_id": task_id},
        )

    return keep



from types import ModuleType
from typing import Any, TypeVar

from rics.misc import format_kwargs, get_by_full_name

T = TypeVar("T")


def initialize(
    name: str,
    kwargs: dict[str, Any],
    instance_of: type[T],
    default_module: ModuleType | None,
) -> T:
    cls = get_by_full_name(name, default_module)  # Or factory function.
    instance = cls(**kwargs)
    if isinstance(instance, instance_of):
        return instance

    msg = (
        f"Expected an instance of '{instance_of.__name__}', but {name}({format_kwargs(kwargs)}) produced: {instance!r}"
    )
    raise TypeError(msg)



"""Data structure IO exceptions."""

from typing import Any as _Any


class DataStructureIOError(TypeError):
    """Base class for IO exceptions."""

    _url = "https://id-translation.readthedocs.io/en/stable/api/id_translation.dio.html#user-defined-integrations"

    def __init__(self, msg: str) -> None:
        super().__init__(msg)
        self.add_note(f"Hint: {self._url}")


class UntranslatableTypeError(DataStructureIOError):
    """Exception indicating that a type cannot be translated.

    Args:
        t: A type.
    """

    def __init__(self, t: type[_Any]) -> None:
        super().__init__(f"Type {t} cannot be translated.")


class NotInplaceTranslatableError(DataStructureIOError):
    """Exception indicating that a type cannot be translated in-place.

    Args:
        arg: Something that can't be translated inplace.
    """

    def __init__(self, arg: _Any) -> None:
        super().__init__(f"Inplace translation not possible or implemented for type: {type(arg)}")



import json
import logging
import sys
from hashlib import sha256
from pathlib import Path
from typing import TYPE_CHECKING, Any

from id_translation.toml._load_toml import load_toml_file

from ._base_metadata import BaseMetadata
from ._metaconf import Metaconf

LOGGER = logging.getLogger(__package__).getChild("Translator").getChild("config")

if TYPE_CHECKING:
    from id_translation import Translator


class ConfigMetadata(BaseMetadata):
    """Metadata pertaining to how a :class:`.Translator` instance was initialized from TOML configuration.

    Equivalence:
        Configs are equivalent if and only if...

        - They have the same top-level dependency versions, and
        - Use the same fully qualified class name, and
        - The main configuration is equal after parsing, and
        - They have the same number of auxiliary (`"extra"`) fetcher configurations, and
        - All auxiliary fetcher configurations are equal after parsing.

    The :class:`Metaconf` is not explicitly included in the equivalence check, but changing it will invalidate cached
    instances. Environment variable changes may also invalidate the cache if :attr:`EnvConf.allow_interpolation` is set
    and interpolations such as ``${VAR}`` are present.

    Args:
        main: Absolute path and fingerprint of the main translation configuration."
        extra_fetchers: Absolute path and fingerprint of configuration files for auxiliary fetchers.
        clazz: String representation of the class type.
        metaconf: A :class:`Metaconf` instance that determines, among other things, how other config paths are processed.
        kwargs: Forwarded to base class.
    """

    def __init__(
        self,
        main: tuple[Path, str],
        extra_fetchers: tuple[tuple[Path, str], ...],
        clazz: str,
        metaconf: Metaconf,
        **kwargs: Any,
    ) -> None:
        if "versions" not in kwargs:
            kwargs["versions"] = {
                "python": metaconf.equivalence.python_version.format(v=sys.version_info),
                **self.get_package_versions(metaconf.equivalence.extra_packages),
            }
        super().__init__(**kwargs)
        self.main = main
        self.extra_fetchers = extra_fetchers
        self.clazz = clazz
        self.metaconf = metaconf

    def _to_dict(self, to_json: dict[str, Any]) -> dict[str, Any]:
        return {
            "main": tuple(map(str, to_json.pop("main"))),
            "extra_fetchers": [tuple(map(str, t)) for t in to_json.pop("extra_fetchers")],
            "metaconf": to_json.pop("metaconf").as_dict(),
            "class": to_json.pop("clazz"),
        }

    @classmethod
    def _deserialize(cls, from_json: dict[str, Any]) -> dict[str, Any]:
        def to_path_tuple(args: list[str]) -> tuple[Path, str]:
            return Path(args[0]), args[1]

        return dict(
            main=to_path_tuple(from_json.pop("main")),
            extra_fetchers=tuple(map(to_path_tuple, from_json.pop("extra_fetchers"))),
            clazz=from_json.pop("class"),
            metaconf=Metaconf.from_dict(from_json.pop("metaconf")),
        )

    def _is_equivalent(self, other: BaseMetadata) -> str:  # pragma: no cover
        assert isinstance(other, ConfigMetadata)  # noqa: S101

        if self.clazz != other.clazz:
            return f"Class not equal. Expected '{self.clazz}', but got '{other.clazz}'"

        if self.main[1] != other.main[1]:
            return f"Main configuration changed. Expected fingerprint {self.main[1]}, but got {other.main[1]}"

        if self.metaconf != other.metaconf:
            return f"Meta configuration changed: {self.metaconf} != {other.metaconf}."

        if len(self.extra_fetchers) != len(other.extra_fetchers):
            return (
                f"Number of auxiliary fetchers changed. Expected {len(self.extra_fetchers)}"
                f" fetchers but got {len(other.extra_fetchers)}"
            )

        for i, (path, fingerprint) in enumerate(self.extra_fetchers):
            _, other_fingerprint = other.extra_fetchers[i]
            if fingerprint != other_fingerprint:
                return (
                    f"Configuration changed for auxiliary fetcher #{i} at {path}. "
                    f"Expected fingerprint {fingerprint}, but got {other_fingerprint}"
                )

        return ""

    @staticmethod
    def from_toml_paths(
        path: str,
        extra_fetchers: list[str],
        clazz: type["Translator[Any, Any, Any]"],
    ) -> "ConfigMetadata":
        """Convenience function for creating ``ConfigMetadata`` instances."""
        metaconf_path = str(Path(path).with_name("metaconf.toml"))
        metaconf = Metaconf.from_path_or_default(metaconf_path)

        def _create_path_tuple(str_path: str) -> tuple[Path, str]:
            p = Path(str_path).expanduser().absolute()
            content = load_toml_file(
                p,
                allow_interpolation=metaconf.env.allow_interpolation,
                allow_nested=metaconf.env.allow_nested,
                allow_blank=metaconf.env.allow_blank,
            )
            return p, sha256(json.dumps(content, sort_keys=True).encode()).hexdigest()

        return ConfigMetadata(
            main=_create_path_tuple(path),
            extra_fetchers=tuple(map(_create_path_tuple, extra_fetchers)),
            clazz=clazz.__module__ + "." + clazz.__qualname__,  # Fully qualified name
            metaconf=metaconf,
        )



"""Mapping errors."""

from typing import Any as _Any


class MappingError(Exception):
    """Base exception class for all mapping-related issues."""

    def __init__(self, msg: str, *, ref: str = "") -> None:
        super().__init__(msg)

        link = "https://id-translation.readthedocs.io/en/stable/documentation/mapping-primer.html"
        if ref:
            link += f"#{ref}"
        self.add_note(f"Hint: See {link} for help.")

        func = "id_translation.logging.enable_verbose_debug_messages"
        if __debug__:
            from id_translation.logging import enable_verbose_debug_messages  # noqa: PLC0415

            expected = enable_verbose_debug_messages.__module__ + "." + enable_verbose_debug_messages.__name__
            assert func == expected  # noqa: S101
        self.add_note(f"Hint: Use `{func}` for detailed output.")


class UnmappedValuesError(MappingError):
    """Raised when there are unmapped values left after filtering and on_unmapped='raise'."""


class ScoringDisabledError(MappingError):
    """Indicates that the scoring logic has been disabled. Raised by :func:`.score_functions.disabled`."""

    def __init__(self, value: _Any, candidates: _Any, context: _Any) -> None:
        super().__init__(
            "Scoring disabled.\n"
            f"The Mapper is working in strict override-only mode, so the {value=} in {context=} "
            f"cannot be mapped to any of the {candidates=}. Possible solutions:\n"
            "    * Add an override or filter for this value, or\n"
            "    * Set strict=False (silently refuse to map instead of raising), or\n"
            "    * Choose an appropriate score function to use.",
            ref="override-only-mapping",
        )
        self.value = value
        self.candidates = candidates
        self.context = context


class AmbiguousScoreError(MappingError):
    """Indicates that the scoring logic has produces ambiguous scores."""

    def __init__(self, kind: str, key: _Any, match0: _Any, match1: _Any, cardinality: str, scores: str) -> None:
        hint = f"\n{scores}\nInspect the matrix above for details. You may wish to use a different scoring method."
        super().__init__(
            f"Ambiguous mapping of {kind}={key!r}; matches ({match0}) and ({match1}) "
            f"are in conflict since {cardinality=}.{hint}"
        )


class UserMappingError(MappingError):
    """A user-defined mapping function did something forbidden."""

    def __init__(self, msg: str, value: _Any, candidates: set[_Any]) -> None:
        super().__init__(msg)
        self.value = value
        self.candidates = candidates


class CardinalityError(MappingError):
    """Base class for cardinality issues."""


class MappingWarning(UserWarning):
    """Base warning class for all mapping-related issues."""


class UnmappedValuesWarning(MappingWarning):
    """Raised when there are unmapped values left after filtering and on_unmapped='raise'."""


class UserMappingWarning(MappingWarning):
    """A user-defined mapping function did something strange."""


class BadFilterError(MappingError):
    """Invalid filter."""



"""Shared API of the :ref:`mapping processes <mapping-primer>` (e.g. names to sources)."""

from ._cardinality import Cardinality
from ._directional_mapping import DirectionalMapping
from ._heuristic_score import HeuristicScore
from ._mapper import Mapper

__all__ = [
    "Cardinality",
    "DirectionalMapping",
    "HeuristicScore",
    "Mapper",
]



import logging
import warnings
from abc import abstractmethod
from collections.abc import Iterable, Sequence
from contextlib import contextmanager
from copy import deepcopy
from time import perf_counter
from typing import Any, Self, final

from rics.collections.dicts import InheritedKeysDict, reverse_dict
from rics.misc import tname
from rics.strings import format_seconds as fmt_sec

from .. import logging as _logging
from ..exceptions import ConnectionStatusError
from ..mapping import HeuristicScore, Mapper
from ..mapping.exceptions import MappingWarning
from ..mapping.score_functions import modified_hamming
from ..offline.types import PlaceholdersTuple, PlaceholderTranslations, SourcePlaceholderTranslations
from ..types import ID, IdType, SourceType
from . import exceptions
from ._cache_access import CacheAccess
from ._fetcher import Fetcher
from .exceptions import CacheAccessNotAvailableError
from .types import FetchInstruction, IdsToFetch


class AbstractFetcher(Fetcher[SourceType, IdType]):
    """Base class for retrieving translations from an external source.

    Args:
        mapper: A :class:`.Mapper` instance used to adapt placeholder names in sources to wanted names, i.e.
            the names of the placeholders that are in the translation :class:`.Format` being used.
        allow_fetch_all: If ``False``, an error will be raised when :meth:`fetch_all` is called.
        selective_fetch_all: If ``True``, fetch only from those :attr:`~.HasSources.sources` that contain the required
            :attr:`~.HasSources.placeholders` (after mapping). May reduce the number of sources retrieved.
        identifiers: A collection of hierarchical identifiers. If given, element zero
            of the `identifiers` is added to the :attr:`logger` name for the fetcher.
        optional: If ``True``, this fetcher may be discarded if source/placeholder-enumeration fails in multi-fetcher
            mode.
        cache_access: A :class:`.CacheAccess` instance. Defaults to a NOOP-implementation (i.e. always fetch new data).
    """

    def __init__(
        self,
        *,
        mapper: Mapper[str, str, SourceType] | None = None,
        allow_fetch_all: bool = True,
        selective_fetch_all: bool = True,
        identifiers: Sequence[str] | None = None,
        optional: bool = False,
        cache_access: CacheAccess[SourceType, IdType] | None = None,
    ) -> None:
        self._mapper: Mapper[str, str, SourceType] = mapper or Mapper(**self.default_mapper_kwargs())
        if self._mapper.on_unmapped == "raise":
            warnings.warn(
                "Using on_unmapped='raise' will treat optional placeholders as "
                "required placeholders during normal operation.",
                category=MappingWarning,
                stacklevel=2,
            )

        self._allow_fetch_all: bool = allow_fetch_all
        self._selective_fetch_all = selective_fetch_all

        identifiers = () if identifiers is None else (*identifiers,)
        logger, mapper_logger = self._configure_loggers(identifiers)
        self.logger = logger
        self._mapper.logger = mapper_logger

        self._identifiers: tuple[str, ...] = identifiers
        self._optional = optional

        self._placeholders: dict[SourceType, list[str]] | None = None

        if cache_access is None:
            cache_access = _NOOP_CACHE_ACCESS
        else:
            cache_access.set_parent(self)
        self._cache_access = cache_access

    @final  # Prevent accidental overriding
    def initialize_sources(self, task_id: int | None = None, *, force: bool = False) -> None:
        if not (self._placeholders is None or force):
            return

        start = perf_counter()
        if task_id is None:
            task_id = _logging.generate_task_id(start)

        logger = self.logger
        if _logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Begin initialization of '{self._cls_name()}' at {hex(id(self))}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.initialize_sources, "enter"),
                ),
            )

        self._placeholders = self._initialize_sources(task_id)
        if self._placeholders is None:
            msg = f"Call to {self._cls_name()}.{self.initialize_sources.__name__}() failed."
            raise RuntimeError(msg)

        if logger.isEnabledFor(logging.INFO):
            seconds = perf_counter() - start
            logger.info(
                f"Finished initialization of '{self._cls_name()}' in {fmt_sec(seconds)}: {self}",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.initialize_sources, "exit"),
                    seconds=seconds,
                ),
            )

    @abstractmethod
    def _initialize_sources(self, task_id: int) -> dict[SourceType, list[str]]:
        """Perform a full (re) discovery of sources and placeholders."""

    @final
    @property
    def selective_fetch_all(self) -> bool:
        """If set, reduce the amount of data fetched by :meth:`fetch_all`."""
        return self._selective_fetch_all

    @final
    @property
    def placeholders(self) -> dict[SourceType, list[str]]:
        if self._placeholders is None:
            self.initialize_sources()
            return self.placeholders

        return self._placeholders

    @final
    @property
    def sources(self) -> list[SourceType]:
        return list(self.placeholders)

    @final
    @property
    def identifiers(self) -> tuple[str, ...]:
        """A collection of hierarchical identifiers for this fetcher."""
        return self._identifiers

    def map_placeholders(
        self,
        source: SourceType,
        placeholders: Iterable[str],
        *,
        candidates: Iterable[str] | None = None,
        task_id: int | None = None,
    ) -> dict[str, str | None]:
        """Map `placeholder` names to the actual names seen in `source`.

        This method calls ``Mapper.apply(values=placeholders, candidates=candidates, context=source)`` using the local
        :attr:`.AbstractFetcher.mapper` instance.

        Args:
            source: The source to map placeholders for.
            placeholders: Desired :attr:`~.Format.placeholders`.
            candidates: A subset of candidates (placeholder names) in `source` to map with `placeholders`.
            task_id: Used for logging.

        Returns:
            A dict ``{wanted_placeholder_name: actual_placeholder_name_in_source}``, where
            `actual_placeholder_name_in_source` will be ``None`` if the wanted placeholder could not be mapped to any of
            the candidates available for the source.

        Raises:
            UnknownSourceError: If `source` is not in :attr:`sources`.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.
        """
        start = perf_counter()

        if self._placeholders is not None and source not in self._placeholders:
            # Check the underlying attribute to avoid infinite recursion in _initialize_sources()-implementations that
            # call this method. This is typically done indirectly via id_column(), which requires explicit candidates.
            raise exceptions.UnknownSourceError({source}, self.sources)
        if candidates is None:
            # May lead to infinite recursion if _initialize_sources() calls map_placeholders(candidates=None).
            candidates = self.placeholders[source]

        candidates = set(candidates)
        placeholders = set(placeholders)

        logger = self.logger
        emit_key_event = self._placeholders or (_logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG))
        if emit_key_event:
            logger.debug(
                f"Begin mapping of wanted {placeholders=} to actual placeholders={candidates} for {source=}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.map_placeholders, "enter"),
                    values=list(placeholders),
                    candidates=list(candidates),
                    context=source,
                ),
            )

        dm = self.mapper.apply(placeholders, candidates, context=source, task_id=task_id)

        ans: dict[str, str | None] = dm.flatten()  # type: ignore[assignment]
        for not_mapped in placeholders.difference(ans):
            ans[not_mapped] = None

        if emit_key_event:
            seconds = perf_counter() - start
            logger.debug(
                f"Finished placeholder mapping for {source=}: {ans}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.map_placeholders, "exit"),
                    seconds=seconds,
                    mapping=ans,
                    context=source,
                ),
            )

        return ans

    def id_column(
        self,
        source: SourceType,
        *,
        candidates: Iterable[str],
        task_id: int | None = None,
    ) -> str | None:
        """Return the ID column for `source`."""
        if not candidates:
            msg = f"Bad {candidates=} argument; must be a non-empty collection."
            raise TypeError(msg)
        return self.map_placeholders(source, [ID], candidates=candidates, task_id=task_id)[ID]

    @property
    def mapper(self) -> Mapper[str, str, SourceType]:
        """Return the ``Mapper`` instance used for placeholder name mapping."""
        return self._mapper

    @property
    def cache_access(self) -> CacheAccess[SourceType, IdType]:
        """Return the :class:`.CacheAccess` for this fetcher."""
        cache_access = self._cache_access

        if cache_access is not _NOOP_CACHE_ACCESS:
            return cache_access

        link = "https://id-translation.readthedocs.io/en/stable/documentation/examples/caching/caching.html"
        msg = f"{self} does not have a `CacheAccess`.\nHint: {link}"
        raise CacheAccessNotAvailableError(msg)

    @property
    def online(self) -> bool:
        return False  # pragma: no cover

    def assert_online(self) -> None:
        """Raise an error if offline.

        Raises:
            ConnectionStatusError: If not online.
        """
        if not self.online:  # pragma: no cover
            raise ConnectionStatusError("disconnected")

    @property
    def allow_fetch_all(self) -> bool:
        return self._allow_fetch_all

    @property
    def logger(self) -> logging.Logger:
        """Return the ``Logger`` that is used by this instance."""
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger) -> None:
        self._logger = logger

    @property
    def optional(self) -> bool:
        return self._optional

    def fetch(
        self,
        ids_to_fetch: Iterable[IdsToFetch[SourceType, IdType]],
        placeholders: Iterable[str] = (),
        required: Iterable[str] = (),
        task_id: int | None = None,
        enable_uuid_heuristics: bool = False,
    ) -> SourcePlaceholderTranslations[SourceType]:
        start = perf_counter()
        if task_id is None:
            task_id = _logging.generate_task_id(start)

        placeholders = tuple(placeholders)
        required_placeholders = set(required)
        ids_to_fetch = tuple(ids_to_fetch)

        logger = self.logger
        if logger.isEnabledFor(logging.DEBUG):
            num_ids = {itf.source: len(itf.ids) for itf in ids_to_fetch}
            logger.debug(
                f"Begin fetching {placeholders=} for {len(num_ids)} sources: {num_ids}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.fetch, "enter"),
                    placeholders=placeholders,
                    required_placeholders=tuple(required_placeholders),
                    num_ids=num_ids,
                    fetch_all=False,
                ),
            )

        rv = {
            itf.source: self._fetch_translations(
                itf.source,
                placeholders,
                required_placeholders=required_placeholders,
                ids=itf.ids,
                task_id=task_id,
                enable_uuid_heuristics=enable_uuid_heuristics,
            )
            for itf in ids_to_fetch
        }

        if logger.isEnabledFor(logging.INFO):
            seconds = perf_counter() - start

            num_ids = {itf.source: len(itf.ids) for itf in ids_to_fetch}
            placeholders_returned = {source: len(pht.placeholders) for source, pht in rv.items()}
            pretty = self._format_fetch_result(rv, num_ids)

            logger.info(
                f"Finished fetching from {len(rv)} sources in {fmt_sec(seconds)}: {pretty}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.fetch, "exit"),
                    seconds=seconds,
                    sources=[*rv],
                    placeholders_returned=placeholders_returned,
                    num_ids=num_ids,
                    num_ids_returned={source: len(pht.records) for source, pht in rv.items()},
                    fetch_all=False,
                ),
            )

        return rv

    def fetch_all(
        self,
        placeholders: Iterable[str] = (),
        *,
        required: Iterable[str] = (),
        sources: set[SourceType] | None = None,
        task_id: int | None = None,
        enable_uuid_heuristics: bool = False,
    ) -> SourcePlaceholderTranslations[SourceType]:
        start = perf_counter()
        if task_id is None:
            task_id = _logging.generate_task_id()

        if not self._allow_fetch_all:
            raise exceptions.ForbiddenOperationError("FETCH_ALL", reason=f"not allowed by {self}.")

        placeholders = tuple(placeholders)
        required_placeholders = set(required)

        logger = self.logger
        if logger.isEnabledFor(logging.DEBUG):
            wanted = self.sources if sources is None else list(sources)
            logger.debug(
                f"Begin fetching all IDs for {placeholders=} for {len(wanted)}/{len(self.sources)}: {wanted}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.fetch_all, "enter"),
                    placeholders=placeholders,
                    required_placeholders=tuple(required_placeholders),
                    wanted_sources=wanted,
                    num_ids=None,
                    fetch_all=True,
                ),
            )

        with self._fetch_all_mapping_context():
            rv = self._fetch_all(
                tuple(placeholders),
                required_placeholders=set(required),
                wanted_sources=sources,
                task_id=task_id,
                enable_uuid_heuristics=enable_uuid_heuristics,
            )

        if logger.isEnabledFor(logging.INFO):
            seconds = perf_counter() - start

            pretty = self._format_fetch_result(rv, None)
            logger.info(
                f"Finished fetching all IDs from {len(self.sources) if sources is None else len(sources)}/"
                f"{len(self.sources)} sources in {fmt_sec(seconds)}: {pretty}.",
                extra=dict(
                    task_id=task_id,
                    event_key=_logging.get_event_key(self.fetch, "exit"),
                    seconds=seconds,
                    sources=[*rv],
                    placeholders_returned={source: len(pht.placeholders) for source, pht in rv.items()},
                    num_ids=None,
                    num_ids_returned={source: len(pht.records) for source, pht in rv.items()},
                    fetch_all=True,
                ),
            )

        return rv

    def _fetch_all(
        self,
        placeholders: PlaceholdersTuple,
        required_placeholders: set[str],
        wanted_sources: set[SourceType] | None,
        task_id: int,
        enable_uuid_heuristics: bool,
    ) -> SourcePlaceholderTranslations[SourceType]:
        if wanted_sources is None:
            wanted_sources = {*self.sources}

        if self._selective_fetch_all:
            # There's nothing stopping us from doing this for regular fetching. But we assume that then the user wants
            # fetching to fail if explicit IDs can't be translated as specified.
            sources = [
                source
                for source in wanted_sources
                if required_placeholders.issubset(self._wanted_to_actual(source, required_placeholders, task_id))
            ]

            if discarded := wanted_sources.difference(sources):
                self.logger.info(
                    f"Ignoring {len(discarded)} sources {discarded} since required "
                    f"placeholders {sorted(required_placeholders)} could not be mapped by {self}.",
                    extra={"task_id": task_id},
                )
        else:
            sources = [*wanted_sources]

        source_translations = {}
        for source in sources:
            translations = self._fetch_translations(
                source,
                placeholders or (*self.placeholders[source],),
                required_placeholders=required_placeholders,
                task_id=task_id,
                enable_uuid_heuristics=enable_uuid_heuristics,
            )
            source_translations[source] = translations

        return source_translations

    @contextmanager
    def _fetch_all_mapping_context(self):  # type: ignore  # noqa
        original_mapper = self._mapper

        on_unmapped = "ignore"
        selective_fetch_all = self._selective_fetch_all
        if not (selective_fetch_all and original_mapper.on_unmapped != on_unmapped):
            yield
            return

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(
                f"Using Mapper.{on_unmapped=} until the current {self.fetch_all.__qualname__}-operation"
                f" finishes, since {selective_fetch_all=}."
            )
        self._mapper = self._mapper.copy(on_unmapped=on_unmapped)
        try:
            yield
        finally:
            self._mapper = original_mapper

    @abstractmethod
    def fetch_translations(self, instr: FetchInstruction[SourceType, IdType]) -> PlaceholderTranslations[SourceType]:
        """Retrieve placeholder translations from the source.

        Args:
            instr: A single :class:`.FetchInstruction` for IDs to fetch. If IDs is ``None``, the fetcher should
                retrieve data for as many IDs as possible.

        Returns:
            Placeholder translation elements.

        Raises:
            UnknownPlaceholderError: If the placeholder is unknown to the fetcher.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.
        """

    def _fetch_translations(
        self,
        source: SourceType,
        placeholders: PlaceholdersTuple,
        *,
        required_placeholders: set[str],
        task_id: int,
        enable_uuid_heuristics: bool,
        ids: set[IdType] | None = None,
    ) -> PlaceholderTranslations[SourceType]:
        placeholders = self._deduplicate(placeholders)
        reverse_mappings, instr = self._make_fetch_instruction(
            source,
            placeholders,
            required_placeholders=required_placeholders,
            ids=ids,
            task_id=task_id,
            enable_uuid_heuristics=enable_uuid_heuristics,
        )

        translations: PlaceholderTranslations[SourceType] | None = None

        cache = self._cache_access
        logger: logging.Logger | None = None
        store_cache = False

        def log_cache(msg: str, event: str) -> None:
            if logger is None:
                return

            pretty = f"{type(cache).__name__}[{source=}]"
            logger.debug(msg.format(pretty), extra={"source": instr.source, "task_id": task_id, "cache_event": event})

        if cache.enabled:
            logger = self.logger
            if not logger.isEnabledFor(logging.DEBUG):
                logger = None

            translations = cache.load(instr)
            store_cache = translations is None

            if logger:
                value = f"{len(translations.records)} IDs" if translations else None
                cache_event = "hit" if translations else "miss"
                log_cache(f"{{}}.load() returned {value}.", cache_event)

        if translations is None:
            translations = self._call_user_impl(instr)

        if reverse_mappings:
            # The mapping is only in reverse from the Fetchers point-of-view; we're mapping back to "proper" values.
            translations.placeholders = tuple(reverse_mappings.get(p, p) for p in translations.placeholders)

        translations.id_pos = translations.placeholders.index(ID)
        translations.placeholder_aliases.update(reverse_mappings)

        available_placeholders = {*translations.placeholders, *translations.placeholder_aliases}
        unmapped_required_placeholders = required_placeholders.difference(available_placeholders)
        if unmapped_required_placeholders:
            self._verify_placeholders(reverse_mappings or {}, source, unmapped_required_placeholders)

        if store_cache:
            log_cache(f"Calling {{}}.store() with {len(translations.records)} IDs.", "store")
            cache.store(instr, translations)

        return translations

    @classmethod
    def _deduplicate(cls, placeholders: PlaceholdersTuple) -> PlaceholdersTuple:
        unique = []  # Hashing is slower for few elements
        for placeholder in placeholders:
            if placeholder not in unique:
                unique.append(placeholder)
        return tuple(unique)

    def _call_user_impl(self, instr: FetchInstruction[SourceType, IdType]) -> PlaceholderTranslations[SourceType]:
        start = perf_counter()

        source = instr.source
        placeholders = instr.placeholders

        logger = self.logger
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Begin fetching {'all' if instr.ids is None else len(instr.ids)} IDs from {source=}. Placeholders: {placeholders}.",
                extra=dict(
                    event_key=_logging.get_event_key(self.fetch_translations, "enter"),
                    source=source,
                    placeholders=instr.placeholders,
                    num_ids=None if instr.ids is None else len(instr.ids),
                    fetch_all=instr.fetch_all,
                    task_id=instr.task_id,
                ),
            )

        translations = self.fetch_translations(instr)

        if logger.isEnabledFor(logging.DEBUG):
            seconds = perf_counter() - start
            logger.debug(
                f"Finished fetching {len(translations.records)} IDs from source='{translations.source}'"
                f" in {fmt_sec(seconds)}. Placeholders: {translations.placeholders}.",
                extra=dict(
                    task_id=instr.task_id,
                    event_key=_logging.get_event_key(self.fetch_translations, "exit"),
                    seconds=seconds,
                    source=source,
                    placeholders=translations.placeholders,
                    num_ids=None if instr.ids is None else len(instr.ids),
                    num_ids_returned=len(translations.records),
                    fetch_all=instr.fetch_all,
                ),
            )

        return translations

    def _verify_placeholders(self, reverse_mappings: dict[str, str], source: SourceType, unmapped: set[str]) -> None:
        hint = ""
        if unmapped.intersection(reverse_mappings.values()):
            r = reverse_dict(reverse_mappings)
            bad_mappings = {b: r[b] for b in unmapped}
            hint = (
                f"\nHint: Mapping {bad_mappings} for required placeholders (keys) were made to placeholders that do not"
                f" exist. The override configuration {self.mapper._overrides} may be incorrect."
            )
        raise exceptions.UnknownPlaceholderError(
            f"Required placeholders {unmapped} not recognized. For {source=}, known placeholders are: "
            f"{sorted(self.placeholders[source])} for {self}.{hint}"
        )

    def _make_fetch_instruction(
        self,
        source: SourceType,
        placeholders: PlaceholdersTuple,
        required_placeholders: set[str],
        ids: set[IdType] | None,
        task_id: int,
        enable_uuid_heuristics: bool,
    ) -> tuple[dict[str, str], FetchInstruction[SourceType, IdType]]:
        required_placeholders.add(ID)
        if ID not in placeholders:
            placeholders = (ID, *placeholders)

        wanted_to_actual = self._wanted_to_actual(source, placeholders, task_id)

        actual_to_wanted = {actual: wanted for wanted, actual in wanted_to_actual.items() if wanted != actual}
        if actual_to_wanted:
            # We'll just map what we can here. If anything is missing it'll be caught later.
            def apply(c: Iterable[str]) -> Iterable[str]:
                return (wanted_to_actual[p] for p in c if p in wanted_to_actual)

            placeholders = tuple(apply(placeholders))
            required_placeholders = set(apply(required_placeholders))

        return (
            actual_to_wanted,
            FetchInstruction(
                source=source,
                placeholders=placeholders,
                required=required_placeholders,
                ids=None if ids is None else set(ids),
                task_id=task_id,
                enable_uuid_heuristics=enable_uuid_heuristics,
            ),
        )

    def _wanted_to_actual(
        self, source: SourceType, wanted_placeholders: Iterable[str], task_id: int | None = None
    ) -> dict[str, str]:
        wanted_to_actual = self.map_placeholders(source, wanted_placeholders, task_id=task_id)
        return {wanted: actual for wanted, actual in wanted_to_actual.items() if actual is not None}

    @classmethod
    def default_mapper_kwargs(cls) -> dict[str, Any]:
        """Return default :class:`.Mapper` arguments for ``AbstractFetcher`` implementations."""
        return dict(
            score_function=HeuristicScore(
                cls.default_score_function,  # type: ignore
                heuristics=[("force_lower_case", {})],
            ),
            overrides=InheritedKeysDict(),
        )

    @classmethod
    def default_score_function(cls, value: str, candidates: Iterable[str], context: str) -> Iterable[float]:
        """Compute score for candidates."""
        return modified_hamming(value, candidates, context)

    def __str__(self) -> str:
        class NoSources:
            def __repr__(self) -> str:
                return "<no sources>"

        sources = self.sources if self.sources else NoSources()
        return f"{type(self).__name__}({sources=})"

    def _cls_name(self) -> str:
        return tname(self, include_module=True).removeprefix(__package__ + ".")

    def __deepcopy__(self, memo: dict[int, Any] = {}) -> Self:  # noqa: B006
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))

        return result

    @classmethod
    def _configure_loggers(cls, identifiers: tuple[str, ...]) -> tuple[logging.Logger, logging.Logger]:
        config_file: str | None = None
        for identifier in identifiers:
            if identifier.endswith(".toml"):
                config_file = identifier
                break

        adapter = _AbstractFetcherLogAdapter(cls.__module__ + "." + cls.__name__, config_file=config_file)

        logger = logging.getLogger(__package__)
        logger.addFilter(adapter)

        mapper_logger = logger.getChild("map")
        mapper_logger.addFilter(adapter)

        return logger, mapper_logger

    @classmethod
    def _format_fetch_result(
        cls,
        source_translation: dict[SourceType, PlaceholderTranslations[SourceType]],
        num_ids_requested: dict[SourceType, int] | None,
    ) -> str:
        def add_requested(source: SourceType) -> str:
            if num_ids_requested is None:
                return ""

            return f"/{num_ids_requested[source]}"

        return ", ".join(
            f"[{source!r} x {pht.placeholders} x {len(pht.records)}{add_requested(source)} IDs]"
            for source, pht in source_translation.items()
        )


class _AbstractFetcherLogAdapter(logging.Filter):
    def __init__(
        self,
        cls: str,
        *,
        config_file: str | None,
    ) -> None:
        super().__init__()
        self.config_file = config_file
        self.cls = cls

    def filter(self, record: logging.LogRecord) -> bool:
        record.fetcher_config_file = self.config_file
        record.fetcher_class = self.cls
        return True


class NoopCacheAccess(CacheAccess[Any, Any]):
    @property
    def enabled(self) -> bool:
        return False

    def _raise(self, *_: Any, **__: Any) -> None:
        raise NotImplementedError

    store = _raise
    load = _raise


_NOOP_CACHE_ACCESS = NoopCacheAccess()



from collections.abc import Sequence
from typing import Any
from uuid import UUID


def try_cast_many(ids: Sequence[Any]) -> Sequence[Any]:
    """Try casting `ids` to UUIDs, falling back to the original input."""
    if len(ids) == 0:
        return ids

    try:
        return list(map(UUID, ids))
    except (ValueError, AttributeError, TypeError):
        return ids


def try_cast_one(idx: Any) -> Any:
    """Try casting `idx` to a UUID, falling back to the original input."""
    if isinstance(idx, UUID):
        return idx

    try:
        return UUID(idx)
    except (ValueError, AttributeError, TypeError):
        return idx



import logging
import warnings
from collections.abc import Collection, Iterable
from copy import deepcopy
from dataclasses import dataclass
from time import perf_counter
from typing import Any, Generic, Literal, Self, TypeAlias, Unpack
from urllib.parse import quote_plus
from uuid import UUID

import sqlalchemy
from rics.misc import format_kwargs
from rics.strings import format_perf_counter as fmt_perf
from sqlalchemy import BINARY, CHAR, TypeDecorator

from .. import _uuid_utils
from .. import logging as _logging
from ..exceptions import ConnectionStatusError
from ..offline.types import PlaceholderTranslations
from ..translator_typing import AbstractFetcherParams
from ..types import ID, IdType
from . import exceptions
from ._abstract_fetcher import AbstractFetcher
from .exceptions import FetcherWarning
from .types import FetchInstruction

BETWEEN_CLAUSE_MIN_ID_COUNT = 100
BETWEEN_CLAUSE_MAX_OVERFETCH_FACTOR = 2.5


@dataclass(frozen=True)
class TableSummary(Generic[IdType]):
    """Brief description of a known table."""

    name: str
    """Name of the table."""
    columns: sqlalchemy.sql.ColumnCollection[str, Any]
    """A flag indicating that the FETCH_ALL-operation is permitted for this table."""
    fetch_all_permitted: bool
    """A flag indicating that the FETCH_ALL-operation is permitted for this table."""
    id_column: sqlalchemy.Column[IdType]
    """The ID column of the table."""


class SqlFetcher(AbstractFetcher[str, IdType]):
    """Fetch data from a SQL source.

    Args:
        connection_string: A SQLAlchemy connection string.
        password: Password to insert into the connection string. Will be escaped to allow for special characters. If
            given, the connection string must contain a password key, eg; ``dialect://user:{password}@host:port``.
        whitelist_tables: The only tables the fetcher may access.
        blacklist_tables: The only tables the fetcher may *not* access.
        schema: Database schema to use. Typically needed only if `schema` is not the default schema for the user
            specified in the connection string.
        include_views: If ``True``, the fetcher will discover and query views as well.
        engine_kwargs: A dict of keyword arguments for :func:`sqlalchemy.create_engine`.
        **kwargs: See :class:`AbstractFetcher`.

    Raises:
        ValueError: If both `whitelist_tables` and `blacklist_tables` are given.

    Notes:
        Inheriting classes may override on or more of the following methods to further customize operation.

        * :meth:`create_engine`; initializes the SQLAlchemy engine. Calls ``parse_connection_string``.
        * :meth:`parse_connection_string`; does basic URL encoding. Called by ``create_engine``.
        * :meth:`select_where`; filter values on the :attr:`~.TableSummary.id_column` of the current table.
        * :meth:`make_table_summary`; creates :class:`TableSummary` instances.
        * :meth:`uuid_like`; determine if :meth:`casting <cast_id_column_to_uuid>` is needed.
        * :meth:`cast_id_column_to_uuid`; attempt to cast the `id_column` to ``UUID``.

        Overriding should be done with care, as methods may call each other internally.
    """

    TableSummary: TypeAlias = TableSummary  # Reexport - part of API/docs.

    def __init__(
        self,
        connection_string: str,
        password: str | None = None,
        whitelist_tables: Iterable[str] | None = None,
        blacklist_tables: Iterable[str] = (),
        schema: str | None = None,
        include_views: bool = False,
        engine_kwargs: dict[str, Any] | None = None,
        **kwargs: Unpack[AbstractFetcherParams[str, IdType]],
    ) -> None:
        super().__init__(**kwargs)
        self._engine_kwargs = engine_kwargs or {}
        self._engine: sqlalchemy.Engine | None = None

        engine, engine_str = self._create_engine(connection_string, password)
        self._engine = engine
        self._estr = engine_str

        self._schema = schema
        self._reflect_views = include_views

        self._blacklist = set(blacklist_tables)

        self._table_summaries: dict[str, TableSummary[IdType]] = {}

        self._whitelist: list[str] | None
        if whitelist_tables is None:
            self._whitelist = None
        else:
            if blacklist_tables:
                raise ValueError("At most one of whitelist and blacklist may be given.")  # pragma: no cover
            self._whitelist = list(set(whitelist_tables))

            if len(self._whitelist) == 0:
                msg = f"Got empty 'whitelist_tables' argument. No tables will be available to {self}."
                self.logger.getChild("sql").warning(msg)
                warnings.warn(msg, category=FetcherWarning, stacklevel=2)

    def _create_engine(
        self,
        connection_string: str,
        password: str | None = None,
    ) -> tuple[sqlalchemy.Engine | None, str]:
        engine: sqlalchemy.Engine | None
        if not self.optional:
            engine = self.create_engine(connection_string, password, self._engine_kwargs)
            return engine, repr(str(engine.url))

        try:
            engine = self.create_engine(connection_string, password, self._engine_kwargs)
            engine_str = repr(str(engine.url))
        except Exception as e:
            engine = None
            engine_str = f"no engine: {e!r}"

        return engine, engine_str

    def select_where(
        self,
        select: sqlalchemy.sql.Select[tuple[IdType, ...]],
        *,
        ids: set[IdType] | None,  # noqa: ARG002
        id_column: sqlalchemy.sql.ColumnElement[IdType],  # noqa: ARG002
        table: sqlalchemy.Table,  # noqa: ARG002
    ) -> sqlalchemy.sql.Select[tuple[IdType, ...]]:
        """User method for modifying SELECT statements.

        The default implementation returns `select` as-is. Selection based on IDs is done before this method is called.
        Users may override this method to change what and which data is returned, e.g. by additional WHERE-clauses.

        Args:
            select: A ``sqlalchemy.sql.Select`` element. If returned as-is, all IDs in the table will be fetched.
            ids: Set of IDs to fetch. Will be ``None`` if :meth:`~AbstractFetcher.fetch_all` was called.
            id_column: The ID ``sqlalchemy.sql.Column`` of the `table`, from which `ids` are fetched.
            table: Table to `select` from.

        Returns:
            The final statement object to use.
        """
        return select

    @classmethod
    def _select_where(
        cls,
        select: sqlalchemy.sql.Select[tuple[IdType, ...]],
        *,
        ids: set[IdType] | None,
        id_column: sqlalchemy.sql.ColumnElement[IdType],
    ) -> sqlalchemy.sql.Select[tuple[IdType, ...]]:
        if ids is None:
            return select

        if not ids:
            return select.where(sqlalchemy.false())

        if len(ids) > BETWEEN_CLAUSE_MIN_ID_COUNT:
            min_id, max_id = min(ids), max(ids)
            try:
                if (max_id - min_id) / len(ids) < BETWEEN_CLAUSE_MAX_OVERFETCH_FACTOR:  # type: ignore[operator]
                    return select.where(id_column.between(min_id, max_id))
            except TypeError:
                pass  # str/UUID

        return select.where(id_column.in_(ids))

    def fetch_translations(self, instr: FetchInstruction[str, IdType]) -> PlaceholderTranslations[str]:
        ts = self._table_summaries[instr.source]

        if instr.fetch_all and not ts.fetch_all_permitted:  # pragma: no cover
            raise exceptions.ForbiddenOperationError("FETCH_ALL", f"disabled for table '{ts.name}'.")

        id_column: sqlalchemy.sql.elements.Cast | sqlalchemy.Column  # type: ignore[type-arg]
        ids_are_uuid_like = instr.enable_uuid_heuristics and self.uuid_like(ts.id_column, instr.ids)
        if ids_are_uuid_like is False:
            id_column = ts.id_column
        else:
            id_column = self.cast_id_column_to_uuid(
                ts.id_column,
                ids_are_uuid_like="unknown" if ids_are_uuid_like is None else True,
            )

        if instr.fetch_all and not self.selective_fetch_all:
            column_names = list(ts.columns.keys())
        else:
            column_names = [name for name in instr.placeholders if name in ts.columns]
        columns = [id_column if name == ts.id_column.name else ts.columns[name] for name in column_names]
        select = sqlalchemy.select(*columns)

        select = self._select_where(select, ids=instr.ids, id_column=id_column)
        select = self.select_where(select, ids=instr.ids, id_column=id_column, table=ts.id_column.table)

        self._log_query(select, logger_extra={"task_id": instr.task_id, "table": instr.source})

        with self.engine.connect() as conn:
            cursor = conn.execute(select)
            records = tuple(map(tuple, cursor))

        return PlaceholderTranslations(instr.source, tuple(cursor.keys()), records)

    def _log_query(
        self,
        select: sqlalchemy.sql.Select[tuple[IdType, ...]],
        logger_extra: dict[str, Any],
        query_length_limit: int = 512,
    ) -> None:
        if not (_logging.ENABLE_VERBOSE_LOGGING and self.logger.isEnabledFor(logging.DEBUG)):
            return

        try:
            query = str(select.compile(self.engine, compile_kwargs={"literal_binds": True}))
            if len(query) > query_length_limit:
                query = query[:query_length_limit]
            self.logger.debug(f"Full SELECT-query using {self._estr}:\n{query}", extra=logger_extra)
        except Exception as e:
            self.logger.debug(
                f"Failed to render full SELECT-query using {self._estr}:\n{select}",
                exc_info=e,
                extra=logger_extra,
            )

    def uuid_like(
        self,
        id_column: sqlalchemy.Column[IdType],
        ids: set[IdType] | None,
    ) -> bool | None:
        """Determine whether `id_column` should be passed to :meth:`cast_id_column_to_uuid`.

        .. note::

           * Will not be called unless :attr:`.Translator.enable_uuid_heuristics` is ``True``.
           * Only ``False`` will bypass calling :meth:`cast_id_column_to_uuid`.

        Return values:
           * ``True``: Attempt to cast using :meth:`cast_id_column_to_uuid` with ``ids_are_uuid_like=True``.
           * ``False``: Do not cast; :meth:`cast_id_column_to_uuid` will **not** be called.
           * ``None``: Attempt to cast using :meth:`cast_id_column_to_uuid` with ``ids_are_uuid_like='unknown'``.

        Args:
            id_column: The ID ``sqlalchemy.sql.Column`` of the table.
            ids: Set of IDs to fetch. Will be ``None`` if :meth:`~AbstractFetcher.fetch_all` was called.

        Returns:
            One of ``True``, ``False`` and ``None``. See above for explanation.
        """
        try:
            python_type = id_column.type.python_type
        except NotImplementedError:
            python_type = None

        if python_type is UUID:
            return True

        if ids:
            first_id = next(iter(ids))
            maybe_uuid = _uuid_utils.try_cast_one(first_id)
            is_uuid = isinstance(maybe_uuid, UUID)
            return is_uuid
        else:
            return None  # Decide solely based on column type.

    def cast_id_column_to_uuid(
        self,
        id_column: sqlalchemy.Column[IdType],
        *,
        ids_are_uuid_like: Literal[True, "unknown"],
    ) -> sqlalchemy.sql.elements.Cast[IdType] | sqlalchemy.Column[IdType]:
        """Apply UUID heuristics to the ID column.

        This function attempts cast the `id_column` to a suitable type by looking at the type of the column and the
        `ids_are_uuid_like`-flag.

        If the column is already UUID-like (as determined by :meth:`get_metadata`), the column is always returned as-is.

        Args:
            id_column: The ID ``sqlalchemy.sql.Column`` of the table.
            ids_are_uuid_like: One of ``True`` and ``'unknown'`` (never ``False``). The latter typically means that
                :meth:`~AbstractFetcher.fetch_all` was called, but could also be a normal "translation" call without IDs.

        Returns:
            The `id_column` with or without a cast applied.
        """
        python_type = id_column.type.python_type
        length = getattr(id_column.type, "length", None)

        uuid_string_lengths = {_String32Uuid.length, _String36Uuid.length}

        if (ids_are_uuid_like is True or length in uuid_string_lengths) and issubclass(python_type, str):
            if self.engine.dialect.name != "mysql":
                return id_column.cast(sqlalchemy.types.Uuid)  # type: ignore[arg-type]

            # MySQL doesn't work even with SQLAlchemy > 2. This seems to be because UUIDs are converted to strings
            # without dashes by the driver (see log output), and databases may contain dashed UUID-like strings.
            return id_column.cast(_String32Uuid if length == _String32Uuid.length else _String36Uuid)  # type: ignore[arg-type]

        if length == _BinaryUuid.length and issubclass(python_type, bytes):
            # Binary-form UUIDs, e.g. MySQL using UUID_TO_BIN and BIN_TO_UUID.
            return id_column.cast(_BinaryUuid)  # type: ignore[arg-type]

        return id_column

    @property
    def online(self) -> bool:
        return self._engine is not None  # pragma: no cover

    def _initialize_sources(self, task_id: int) -> dict[str, list[str]]:
        if self._engine is None:
            raise ConnectionStatusError(f"disconnected: {self._estr}")

        self._table_summaries = self._get_summaries(task_id)
        return {
            table_name: [str(c.name) for c in table_summary.columns]  # table_name = source
            for table_name, table_summary in self._table_summaries.items()
        }

    @property
    def allow_fetch_all(self) -> bool:
        self.initialize_sources()  # Ensure self._table_summaries is populated.
        return super().allow_fetch_all and all(s.fetch_all_permitted for s in self._table_summaries.values())

    def __str__(self) -> str:
        disconnected = "<disconnected>: " if not self.online else ""

        kwargs: dict[str, Any] = {}
        if self._schema:
            kwargs["schema"] = self._schema
        if self._whitelist:
            kwargs["whitelist"] = self._whitelist
        elif self._table_summaries:
            kwargs["sources"] = [*self._table_summaries]

        pretty_kwargs = ", " + format_kwargs(kwargs, max_value_length=1024) if kwargs else ""
        return f"{type(self).__name__}({disconnected}{self._estr}{pretty_kwargs})"

    @property
    def engine(self) -> sqlalchemy.engine.Engine:
        """The :class:`~sqlalchemy.engine.Engine` used by this fetcher."""
        self.assert_online()
        assert self._engine is not None  # noqa: S101
        return self._engine

    def close(self) -> None:
        """Close the fetcher, discarding the :attr:`engine`."""
        if self._engine is None:
            return

        self.logger.debug("Dispose %s.", self._estr)
        self._table_summaries = {}
        self._engine.dispose()
        self._engine = None

    @classmethod
    def create_engine(
        cls,
        connection_string: str,
        password: str | None,
        engine_kwargs: dict[str, Any],
    ) -> sqlalchemy.engine.Engine:
        """Factory method used by ``__init__``.

        For a more detailed description of the arguments and the behaviour of this function, see the
        :class:`class docstring <SqlFetcher>`.

        Args:
            connection_string: A SQLAlchemy connection string.
            password: Password to insert into the connection string.
            engine_kwargs: A dict of keyword arguments for :func:`sqlalchemy.create_engine`.

        Returns:
            A new ``Engine``.
        """
        return sqlalchemy.create_engine(
            cls.parse_connection_string(connection_string, password),
            **engine_kwargs,
        )

    @classmethod
    def parse_connection_string(cls, connection_string: str, password: str | None) -> str:
        """Parse a connection string."""
        if password:
            if "{password}" in connection_string:
                connection_string = connection_string.format(password=quote_plus(password))
            else:
                warnings.warn(
                    "A password was specified, but the connection string does not have a {password} key.",
                    stacklevel=3,
                )
        return connection_string

    def _get_summaries(self, task_id: int) -> dict[str, TableSummary[IdType]]:
        start = perf_counter()
        metadata = self.get_metadata()

        logger = self.logger
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Metadata for {self._estr} created in {fmt_perf(start)}.",
                extra={"task_id": task_id},
            )

        table_names = {t.name for t in metadata.tables.values()}
        tables: Collection[str]
        if self._whitelist:
            tables = self._whitelist
        else:
            tables = table_names.difference(self._blacklist) if self._blacklist else table_names

        if not tables:  # pragma: no cover
            extra = ""
            if self._whitelist:
                extra = f" (whitelist: {self._whitelist})"
            elif self._blacklist:
                extra = f" (blacklist: {self._blacklist})"

            logger.warning(
                f"{self._estr}: No sources found{extra}. Available tables: {table_names}",
                extra={"task_id": task_id},
            )
            return {}

        discarded = []
        ans = {}
        for name in tables:
            qualified_name = name if self._schema is None else f"{self._schema}.{name}"
            table = metadata.tables[qualified_name]
            id_column_name = self.id_column(table.name, candidates=(c.name for c in table.columns), task_id=task_id)
            if id_column_name is None or (id_column := table.columns.get(id_column_name)) is None:
                reason = self._handle_unknown_table(id_column_name, table, qualified_name)
                if reason:
                    discarded.append((qualified_name, reason))
            else:
                ans[str(name)] = self.make_table_summary(table, id_column)

        if discarded and _logging.ENABLE_VERBOSE_LOGGING and self.logger.isEnabledFor(logging.DEBUG):
            per_reason: dict[str, list[str]] = {}
            for qualified_name, reason in discarded:
                per_reason.setdefault(reason, []).append(qualified_name)
            self.logger.debug(
                f"Discarded {len(discarded)} tables. Reason: {per_reason}",
                extra={"task_id": task_id, "reason": per_reason},
            )

        return ans

    def _handle_unknown_table(
        self,
        id_column: str | None,
        table: sqlalchemy.Table,
        table_name: str,
    ) -> str | None:
        whitelisted = False if self._whitelist is None else table.name in self._whitelist
        unmapped = id_column is None

        if unmapped and not whitelisted:
            return "no ID column"

        messages = []
        if whitelisted:
            messages.append("Misconfigured whitelist table.")
        messages.append(
            f"No suitable ID column found for the table {table_name!r}. "
            f"Known columns: {sorted(c.name for c in table.columns)}."
        )
        if not unmapped:
            messages.append(
                f"This is likely caused by a bad override. "
                f"Update or remove the override {ID!r} -> {id_column!r} from your mapping configuration."
            )
        raise exceptions.UnknownPlaceholderError(" ".join(messages))

    def make_table_summary(
        self,
        table: sqlalchemy.Table,
        id_column: sqlalchemy.Column[IdType],
    ) -> TableSummary[IdType]:
        """Create a table summary.

        This function is called as a part of the fetcher initialization process.

        Args:
            table: The table (source) which is currently being processed.
            id_column: The ID column of `table`

        Returns:
            A summary object for `table`.
        """
        return TableSummary(
            name=str(table.name),
            columns=table.columns.as_readonly(),
            fetch_all_permitted=True,
            id_column=id_column,
        )

    def get_metadata(self) -> sqlalchemy.MetaData:
        """Create a populated metadata object."""
        metadata = sqlalchemy.MetaData(schema=self._schema)
        metadata.reflect(self.engine, only=self._whitelist, views=self._reflect_views)
        return metadata

    def __deepcopy__(self, memo: dict[int, Any] = {}) -> Self:  # noqa: B006
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        for k, v in self.__dict__.items():
            setattr(
                result,
                k,
                self._copy_engine(memo, v, self._engine_kwargs)
                if isinstance(v, sqlalchemy.Engine)
                else deepcopy(v, memo),
            )

        return result

    @staticmethod
    def _copy_engine(memo: dict[int, Any], engine: sqlalchemy.Engine, kwargs: dict[str, Any]) -> sqlalchemy.Engine:
        engine_id = id(engine)

        if engine_id in memo:
            return memo[engine_id]  # type: ignore[no-any-return]

        result = sqlalchemy.create_engine(engine.url, **kwargs)
        memo[engine_id] = result
        return result


SqlFetcher.TableSummary = TableSummary  # Reexport


class _BinaryUuid(TypeDecorator[UUID]):
    length: int = 16
    impl = BINARY(length)
    cache_ok = True

    def process_bind_param(self, value: UUID | None, _dialect: Any) -> bytes | None:
        """Mimic UUID_TO_BIN."""
        return None if value is None else value.bytes

    def process_result_value(self, value: bytes | None, _dialect: Any) -> UUID | None:
        """Mimic BIN_TO_UUID."""
        return None if value is None else UUID(bytes=value)


class _String32Uuid(TypeDecorator[UUID]):
    length: int = 32
    impl = CHAR(length)
    cache_ok = True

    def process_bind_param(self, value: UUID | None, _dialect: Any) -> str | None:
        """To string representation without dashes."""
        return None if value is None else value.hex

    def process_result_value(self, value: str | None, _dialect: Any) -> UUID | None:
        """From string representation without dashes."""
        return None if value is None else UUID(value)


class _String36Uuid(TypeDecorator[UUID]):
    length: int = 36
    impl = CHAR(length)
    cache_ok = True

    def process_bind_param(self, value: UUID | None, _dialect: Any) -> str | None:
        """To string representation with dashes."""
        return None if value is None else str(value)

    def process_result_value(self, value: str | None, _dialect: Any) -> UUID | None:
        """From string representation with dashes."""
        return None if value is None else UUID(value)



from collections.abc import Sequence
from typing import Any

from rics.collections.misc import as_list
from rics.logs import disable_temporarily

from id_translation.offline import TranslationMap
from id_translation.types import IdType, NameType, SourceType

from .._data_structure_io import DataStructureIO


class DictIO(DataStructureIO[dict[NameType, IdType], NameType, SourceType, IdType]):
    """IO implementation for ``dict`` types."""

    priority = 1000

    @classmethod
    def handles_type(cls, arg: Any) -> bool:
        return isinstance(arg, dict)

    @classmethod
    def names(cls, translatable: dict[NameType, IdType]) -> list[NameType]:
        return list(translatable)

    @classmethod
    def extract(cls, translatable: dict[NameType, IdType], names: list[NameType]) -> dict[NameType, Sequence[IdType]]:
        return {name: as_list(translatable[name]) for name in names}

    @classmethod
    def insert(
        cls,
        translatable: dict[NameType, IdType] | dict[NameType, Sequence[IdType]],
        names: list[NameType],
        tmap: TranslationMap[NameType, SourceType, IdType],
        copy: bool,
    ) -> dict[NameType, Any] | None:
        from .._resolve import LOGGER as RESOLVE_IO_LOGGER  # noqa: PLC0415
        from .._resolve import resolve_io  # noqa: PLC0415

        with disable_temporarily(RESOLVE_IO_LOGGER):
            translated = {}
            for key, value in translatable.items():
                dio: DataStructureIO[IdType | Sequence[IdType], NameType, SourceType, IdType] = resolve_io(value)
                if key in names:
                    translated[key] = dio.insert(value, [key], tmap, copy=True)
                else:
                    translated[key] = value

        if copy:
            return translated

        translatable.clear()
        translatable.update(translated)  # type: ignore[arg-type]
        return None



"""General errors for the translation suite."""

from warnings import simplefilter as _simplefilter


class ConfigurationError(TypeError):
    """Raised in case of bad configuration."""


class ConfigurationChangedError(Exception):
    """See :meth:`.Translator.load_persistent_instance`."""


class ConnectionStatusError(ConnectionError):
    """Raised when trying to perform operations in a bad online/offline state."""


class TranslationError(Exception):
    """Base class for translation errors."""


class MissingNamesError(TranslationError):
    """Raised if names could not be derived based on the data type, and aren't explicitly given."""


class TooManyFailedTranslationsError(TranslationError):
    """Raised if too many IDs fail to translate."""


class TranslationWarning(UserWarning):
    """Base class for translation warnings."""


class TranslationDisabledWarning(TranslationWarning):
    """Translation is globally disabled."""


_simplefilter("once", category=TranslationDisabledWarning)



import logging
import warnings
from typing import TYPE_CHECKING, Generic

from rics.collections.misc import as_list

from ..mapping.exceptions import MappingWarning

if TYPE_CHECKING:
    from .._translator import Translator

from ..exceptions import MissingNamesError
from ..mapping.types import UserOverrideFunction
from ..types import IdType, Names, NameToSource, NameType, NameTypes, SourceType, Translatable
from ._base_task import BaseTask

LOGGER = logging.getLogger("id_translation.Translator")


class NamesTask(BaseTask[NameType, SourceType, IdType]):
    """Internal type; not part of the public API."""

    def __init__(
        self,
        caller: "Translator[NameType, SourceType, IdType]",
        translatable: Translatable[NameType, IdType],
        names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None = None,
        ignore_names: Names[NameType] | None = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] | None = None,
        task_id: int | None = None,
    ) -> None:
        super().__init__(caller, translatable, task_id=task_id)

        if not (names is None or ignore_names is None):
            raise ValueError(
                f"Required {names=} cannot be used with {ignore_names=}."
                f"\nHint: Set names=None to use automatic name extraction with user-defined ignored names."
            )

        names, override_function = _handle_input_names(names, override_function)

        # Explicit, user-given names
        self.names_from_user: list[NameType] | None = None if names is None else as_list(names)

        # Names to exclude. May not be combined with explicit names (verified above).
        self.ignore_names: Names[NameType] | None = None
        if ignore_names is not None:
            self.ignore_names = ignore_names if callable(ignore_names) else as_list(ignore_names)

        # A callable (NameType) -> bool, if given
        self.override_function = override_function

        # Task outputs
        self._need_name_extraction: bool = True
        self._extracted_names: list[NameType] | None = None
        self._mapper_input_names: list[NameType] | None = None

    @property
    def extracted_names(self) -> list[NameType]:
        """Names extracted from `translatable`, based on ``type(translatable)``."""
        if self._extracted_names is not None:
            return self._extracted_names

        if not self._need_name_extraction:
            # Users should never see this. If they do, something went (very) wrong internally.
            raise MissingNamesError(
                "You're not supposed to see this. Please report this issue at:\n"
                "https://github.com/rsundqvist/id-translation/issues/new"
            )
        self._need_name_extraction = False

        names: list[NameType] | None = self.io.names(self.translatable)
        if names is None:
            raise MissingNamesError(
                f"Failed to derive names for {self.type_name}-type data."
                "\nHint: Use the 'names'-argument to specify names to translate."
            )
        self._extracted_names = names
        return self._extracted_names

    @property
    def mapper_input_names(self) -> list[NameType]:
        """Names that should be mapped to sources."""
        if self._mapper_input_names is None:
            self._mapper_input_names = self._filter_names()
        return self._mapper_input_names

    def _filter_names(self) -> list[NameType]:
        names = self.names_from_user

        if names is not None:
            return as_list(names)

        ignore_names = self.ignore_names  # May not be combined with explicit names; checked earlier.
        names = self.extracted_names
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                f"Name extraction complete. Found {names=} for {self.type_name}-type data.",
                extra={"task_id": self.task_id},
            )

        if names and ignore_names is not None:
            predicate = ignore_names if callable(ignore_names) else set(as_list(ignore_names)).__contains__
            names = [name for name in names if not predicate(name)]

            if not names:
                warnings.warn(
                    f"No names left to translate. All derived names={self.extracted_names} in the {self.type_name}-type"
                    f" data where removed by ignore_names={ignore_names}.",
                    category=MappingWarning,
                    stacklevel=3,
                )

        return names


def _handle_input_names(
    names: NameTypes[NameType] | NameToSource[NameType, SourceType] | None,
    override_function: UserOverrideFunction[NameType, SourceType, None] | None,
) -> tuple[list[NameType] | None, UserOverrideFunction[NameType, SourceType, None] | None]:
    if names is None:
        return None, override_function

    if isinstance(names, dict):
        if override_function is not None:
            raise ValueError(f"Dict-type {names=} cannot be combined with {override_function=}.")

        override_function = _UserDefinedNameToSourceMapping(dict(names))

    return as_list(names), override_function


class _UserDefinedNameToSourceMapping(Generic[NameType, SourceType]):
    def __init__(self, name_to_source: NameToSource[NameType, SourceType]) -> None:
        for name, source in name_to_source.items():
            if source is None:
                raise ValueError(
                    f"Bad name-to-source mapping: {name!r} -> {source!r}."
                    f"\nHint: Remove None-values from names={name_to_source}."
                )
        self._name_to_source = name_to_source

    def __call__(self, name: NameType, _sources: set[SourceType], _context: None) -> SourceType | None:
        return self._name_to_source.get(name)

    def __repr__(self) -> str:
        return f"UserArgument(names={self._name_to_source})"



"""Errors and warnings related to fetching."""

from collections.abc import Iterable as _Iterable
from typing import Any as _Any


class FetcherWarning(RuntimeWarning):
    """Base class for ``Fetcher`` warnings."""


class FetcherError(RuntimeError):
    """Base class for ``Fetcher`` exceptions."""


class ForbiddenOperationError(FetcherError):
    """Exception indicating that the ``Fetcher`` does not support an operation."""

    def __init__(self, operation: str, reason: str) -> None:
        super().__init__(f"Operation '{operation}' " + reason)
        self.operation = operation


class ImplementationError(FetcherError):
    """An underlying implementation did something wrong."""


class UnknownPlaceholderError(FetcherError):
    """Caller requested unknown placeholder name(s)."""


class UnknownIdError(FetcherError):
    """Caller requested unknown id(s)."""


class UnknownSourceError(FetcherError):
    """Caller requested unknown source(s).

    Args:
        unknown_sources: The sources which are not known to the Fetcher.
        sources: Sources known to the fetcher.
        msg: A format string that takes `unknown_sources` and `sources`.
    """

    def __init__(
        self,
        unknown_sources: _Iterable[_Any],
        sources: _Iterable[_Any],
        msg: str = "Sources {unknown_sources} not recognized. Known sources: {sources}.",
    ) -> None:
        self.sources = set(sources)
        self.unknown_sources = set(unknown_sources)
        super().__init__(msg.format(unknown_sources=self.unknown_sources, sources=self.sources))


class DuplicateSourceWarning(FetcherWarning):
    """Duplicate sources detected."""


class DuplicateSourceError(FetcherError):
    """Multiple translations for the same source received."""


class CacheAccessNotAvailableError(FetcherError):
    """Raised when calling :attr:`.AbstractFetcher.cache_access` on an instance that is not cached."""
